----------------------------------------------------------------------------
-- Module Name:        DPE_IASW - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    Provides nitialisation of IASW.
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
--    Procedure INITIALISE is the very first piece of IASW code executed.
--    By calls of procedures CONSTRUCTOR of differente ADA packages it
--    preinitialises packages.  Next all seven tasks of IASW are created,
--    some of them are active immediatelly, others will be activatet later.
----------------------------------------------------------------------------

with CSSW_COMMON_TYPES;  use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;      use CSSW_IF_TYPES;
with DPE_CSSW;
with IASW_COMMON;
with IASW_MAIN;
with IASW_TM;
with IASW_MRTU;
with IASW_REQ_OUT;
with IASW_HK;
with IASW_SC_OUT;
with IASW_HSL;
with IASW_TIME;
with IASW_DFEE;
with IASW_MODES;
with IASW_STATE;
with IASW_GF;
with IASW_BUFFER;
with IASW_PATCHES;
with IASW_STACK1;
with IASW_STACK2;
with IASW_STACK3;
with IASW_STACK4;
with IASW_STACK5;
with IASW_STACK7;
with IASW_STACK8;

package body DPE_IASW is

   package CIT renames CSSW_IF_TYPES;

   procedure TASK_MAIN_STARTER is
   begin
      loop
         begin
            IASW_MAIN.TASK_MAIN;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_MAIN_STARTER;

   procedure TASK_TC_STARTER is
   begin
      loop
         begin
            IASW_MAIN.TASK_TC;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_TC_STARTER;

   procedure TASK_HSL_STARTER is
   begin
      loop
         begin
            IASW_HSL.TASK_HSL;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_HSL_STARTER;

   procedure TASK_SC_OUT_STARTER is
   begin
      loop
         begin
            IASW_SC_OUT.TASK_SC_OUT;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_SC_OUT_STARTER;

   procedure TASK_HK_STARTER is
   begin
      loop
         begin
            IASW_HK.TASK_HK;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_HK_STARTER;

   procedure TASK_REQ_OUT_STARTER is
   begin
      loop
         begin
            IASW_REQ_OUT.TASK_REQ_OUT;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_REQ_OUT_STARTER;

   procedure TASK_STATE_STARTER is
   begin
      loop
         begin
            IASW_STATE.TASK_STATE;
         exception
            when others =>
               null;
         end;
      end loop;
   end TASK_STATE_STARTER;


   procedure INITIALISE is
   begin
      IASW_COMMON.  CONSTRUCTOR;
      IASW_MAIN.    CONSTRUCTOR;
      IASW_TM.      CONSTRUCTOR;
      IASW_MRTU.    CONSTRUCTOR;
      IASW_REQ_OUT. CONSTRUCTOR;
      IASW_HK.      CONSTRUCTOR;
      IASW_SC_OUT.  CONSTRUCTOR;
      IASW_HSL.     CONSTRUCTOR;
      IASW_TIME.    CONSTRUCTOR;
      IASW_DFEE.    CONSTRUCTOR;
      IASW_MODES.   CONSTRUCTOR;
      IASW_STATE.   CONSTRUCTOR;
      IASW_PATCHES. CONSTRUCTOR;
      IASW_GF.      CONSTRUCTOR;
      IASW_BUFFER.  CONSTRUCTOR;

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_MAIN_ID,
         PRIORITY      => IASW_COMMON.TASK_MAIN_PRIORITY,
         START         => TASK_MAIN_STARTER'address,
         STACK         => IASW_STACK1.TASK_MAIN_STACK,
         INITIAL_STATE => CIT.STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_TC_ID,
         PRIORITY      => IASW_COMMON.TASK_TC_PRIORITY,
         START         => TASK_TC_STARTER'address,
         STACK         => IASW_STACK2.TASK_TC_STACK,
         INITIAL_STATE => CIT.STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_HSL_ID,
         PRIORITY      => IASW_COMMON.TASK_HSL_PRIORITY,
         START         => TASK_HSL_STARTER'address,
         STACK         => IASW_STACK3.TASK_HSL_STACK,
         INITIAL_STATE => CIT.STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_SC_OUT_ID,
         PRIORITY      => IASW_COMMON.TASK_SC_OUT_PRIORITY,
         START         => TASK_SC_OUT_STARTER'address,
         STACK         => IASW_STACK4.TASK_SC_OUT_STACK,
         INITIAL_STATE => CIT.STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_HK_ID,
         PRIORITY      => IASW_COMMON.TASK_HK_PRIORITY,
         START         => TASK_HK_STARTER'address,
         STACK         => IASW_STACK5.TASK_HK_STACK,
         INITIAL_STATE => CIT.NOT_STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_REQ_OUT_ID,
         PRIORITY      => IASW_COMMON.TASK_REQ_OUT_PRIORITY,
         START         => TASK_REQ_OUT_STARTER'address,
         STACK         => IASW_STACK7.TASK_REQ_OUT_STACK,
         INITIAL_STATE => CIT.NOT_STARTED );

      DPE_CSSW.INIT_TASK (
         ID            => IASW_COMMON.TASK_STATE_ID,
         PRIORITY      => IASW_COMMON.TASK_STATE_PRIORITY,
         START         => TASK_STATE_STARTER'address,
         STACK         => IASW_STACK8.TASK_STATE_STACK,
         INITIAL_STATE => CIT.NOT_STARTED );

   end INITIALISE;

end DPE_IASW;
