----------------------------------------------------------------------------
-- Module Name:        IASW_BUFFER - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    Provides access to the big buffer of events in RAM.
-- Interfaces:
--    (Define the inputs and outputs)
--    The buffer consists of ten sub-buffers, 30000 wods each.  This is
--    due to memory organization in 1750A.
-- Dependencies:
--    (Describe the preconditions for using this component)
--    Before use the buffer needs to be initialised with use of
--    procedure SET_EVENT_SIZE.
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
--    This package allows acces to the buffer as a whole.  Mostly used is
--    sequencial access with use of procedures PUT_EVENT and GET_EVENT -
--    in this case the buffer works as FIFO.
--    Random (indexed) access is also possible wit use of procedure
--    PEEK_EVENT.  Before use the buffer needs to be initialised.
--    Procedure SET_EVENT_SIZE performs all necessary initialisation.
----------------------------------------------------------------------------

with CSSW_COMMON_TYPES; use CSSW_COMMON_TYPES;
with IASW_COMMON;
with DPE_CSSW;
with IASW_SUBBUFFER_1;
with IASW_SUBBUFFER_2;
with IASW_SUBBUFFER_3;
with IASW_SUBBUFFER_4;
with IASW_SUBBUFFER_5;
with IASW_SUBBUFFER_6;
with IASW_SUBBUFFER_7;
with IASW_SUBBUFFER_8;
with IASW_SUBBUFFER_9;
with IASW_SUBBUFFER_10;

package body IASW_BUFFER is

   procedure PREPARE_BUFFER is   -- local
   begin
      -- Calculate sizes of all sub-buffers.
      SUBBUFFER_SIZES( 1) := IASW_SUBBUFFER_1. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 2) := IASW_SUBBUFFER_2. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 3) := IASW_SUBBUFFER_3. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 4) := IASW_SUBBUFFER_4. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 5) := IASW_SUBBUFFER_5. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 6) := IASW_SUBBUFFER_6. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 7) := IASW_SUBBUFFER_7. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 8) := IASW_SUBBUFFER_8. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES( 9) := IASW_SUBBUFFER_9. GET_SUBBUFFER_SIZE (EVENT_SIZE);
      SUBBUFFER_SIZES(10) := IASW_SUBBUFFER_10.GET_SUBBUFFER_SIZE (EVENT_SIZE);

      -- Calculate total buffer size.
      TOTAL_BUFFER_SIZE := 0;
      for I in 1 .. K_NBUF loop
         TOTAL_BUFFER_SIZE := TOTAL_BUFFER_SIZE + SUBBUFFER_SIZES (I);
         SUBBUFFER_SIZES (I) := TOTAL_BUFFER_SIZE;
      end loop;

      CLEAR_BUFFER;
      ENABLE_INPUT;
   end PREPARE_BUFFER;


   procedure SET_EVENT_SIZE (EVENT_SIZE : in IASW_COMMON.T_EVENT_SIZE) is
   begin
      IASW_BUFFER.EVENT_SIZE := EVENT_SIZE;
      PREPARE_BUFFER;
   end SET_EVENT_SIZE;


   procedure CLEAR_BUFFER is
   begin
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(BUFFER_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      -- Initialize indices.
      READING_INDEX := 0;
      WRITING_INDEX := 0;
      AMOUNT_OF_EVENTS := 0;

      BUFFER_SEMAPHORE :=0;
   end CLEAR_BUFFER;


   procedure ACCESS_EVENT(   -- local
   --
   -- Stores given event in buffer or retrieves it depending on STORING.
   -- INDEX is a place where event should be stored to/retrieved from.
   --
      EVENT : in out IASW_COMMON.T_EVENT;
      INDEX : in T_BUFFER_INDEX;
      STORING : in BOOLEAN
   ) is
   begin
pragma Suppress(all_checks);
      if INDEX < SUBBUFFER_SIZES (1) then
         IASW_SUBBUFFER_1.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX,
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (2) then
         IASW_SUBBUFFER_2.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (1),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (3) then
         IASW_SUBBUFFER_3.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (2),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (4) then
         IASW_SUBBUFFER_4.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (3),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (5) then
         IASW_SUBBUFFER_5.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (4),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (6) then
         IASW_SUBBUFFER_6.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (5),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (7) then
         IASW_SUBBUFFER_7.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (6),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (8) then
         IASW_SUBBUFFER_8.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (7),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (9) then
         IASW_SUBBUFFER_9.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (8),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      elsif INDEX < SUBBUFFER_SIZES (10) then
         IASW_SUBBUFFER_10.ACCESS_EVENT(
            EVENT => EVENT,
            INDEX => INDEX - SUBBUFFER_SIZES (9),
            EVENT_SIZE => EVENT_SIZE,
            STORING => STORING);
         return;
      else
         raise EX_INDEX_OUT_OF_RANGE;
      end if;
   end ACCESS_EVENT;


   procedure PUT_EVENT (EVENT : in out IASW_COMMON.T_EVENT) is
   --
   -- Appends given EVENT at the end of buffer.  Updates indices.
   --
   begin
pragma Suppress(all_checks);
      if IS_BUFFER_FULL then
         raise EX_BUFFER_FULL;
      end if;

      if IS_BUFFER_LOCKED then
         -- raise EX_BUFFER_LOCKED
	 return;
      end if;

      ACCESS_EVENT(
         EVENT => EVENT,
         INDEX => WRITING_INDEX,
         STORING => TRUE);

      -- Update index.

      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(BUFFER_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      WRITING_INDEX := (WRITING_INDEX + 1) mod TOTAL_BUFFER_SIZE;
      AMOUNT_OF_EVENTS := AMOUNT_OF_EVENTS + 1;

      BUFFER_SEMAPHORE := 0;
   end PUT_EVENT;


   procedure GET_EVENT (EVENT : in out IASW_COMMON.T_EVENT) is
   --
   -- Returns EVENT from the top of the buffer.  Updates indices.
   --
   begin
pragma Suppress(all_checks);
      if IS_BUFFER_EMPTY then
         raise EX_BUFFER_EMPTY;
      end if;

      ACCESS_EVENT(
         EVENT => EVENT,
         INDEX => READING_INDEX,
         STORING => FALSE);

      -- Update index.
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(BUFFER_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      AMOUNT_OF_EVENTS := AMOUNT_OF_EVENTS - 1;
      READING_INDEX := (READING_INDEX + 1) mod TOTAL_BUFFER_SIZE;

      BUFFER_SEMAPHORE := 0;
   end GET_EVENT;


   procedure PEEK_EVENT(
   --
   -- Retrieves event of given index without removing it from buffer.
   -- INDEX is counted from the beginning of the buffer (i.e. INDEX = 0
   -- means the event which would be returned by next call of GET_EVENT).
   --
      INDEX : in T_BUFFER_INDEX;
      EVENT : in out IASW_COMMON.T_EVENT
   ) is
   begin
pragma Suppress(all_checks);
      if IS_BUFFER_EMPTY then
         raise EX_BUFFER_EMPTY;
      end if;

      if INDEX >= AMOUNT_OF_EVENTS then
         raise EX_INDEX_OUT_OF_RANGE;
      end if;

      ACCESS_EVENT(
         EVENT => EVENT,
         INDEX => ((READING_INDEX + INDEX) mod TOTAL_BUFFER_SIZE),
         STORING => FALSE);
   end PEEK_EVENT;


   procedure DISCARD_EVENTS (AMOUNT : in T_BUFFER_INDEX) is
   --
   -- Discards given amount of events from the beginning of buffer.
   --
   begin
      if AMOUNT_OF_EVENTS < AMOUNT then
         raise EX_BUFFER_EMPTY;
      end if;

      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(BUFFER_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      AMOUNT_OF_EVENTS := AMOUNT_OF_EVENTS - AMOUNT;
      READING_INDEX := (READING_INDEX + AMOUNT) mod TOTAL_BUFFER_SIZE;
      -- Warning! The above '+' is not safe if AMOUNT > 5000, because
      -- READING_INDEX can be up to 59999.  In practice AMOUNT <= 2400.

      BUFFER_SEMAPHORE := 0;
   end DISCARD_EVENTS;


   function IS_BUFFER_FULL return BOOLEAN is
   begin
      return AMOUNT_OF_EVENTS = TOTAL_BUFFER_SIZE;
   end IS_BUFFER_FULL;


   function IS_BUFFER_EMPTY return BOOLEAN is
   begin
      return AMOUNT_OF_EVENTS = 0;
   end IS_BUFFER_EMPTY;


   function GET_AMOUNT_OF_EVENTS return T_BUFFER_INDEX is
   begin
      return AMOUNT_OF_EVENTS;
   end GET_AMOUNT_OF_EVENTS;


   function GET_TOTAL_BUFFER_SIZE return T_BUFFER_INDEX is
   begin
      return TOTAL_BUFFER_SIZE;
   end GET_TOTAL_BUFFER_SIZE;


   procedure ENABLE_INPUT is
   begin
      BUFFER_LOCKED := FALSE;
   end ENABLE_INPUT;


   procedure DISABLE_INPUT is
   begin
      BUFFER_LOCKED := TRUE;
   end DISABLE_INPUT;


   function IS_BUFFER_LOCKED return BOOLEAN is
   begin
      return BUFFER_LOCKED;
   end IS_BUFFER_LOCKED;


   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE) is
   begin
      DS.AMOUNT_OF_EVENTS := AMOUNT_OF_EVENTS;
   end GET_DPE_STATE;


   procedure GET_HK_DIAG( FLAGS : out IASW_COMMON.T_16_BITS ) is
   begin
      FLAGS(11) := BUFFER_LOCKED;
      FLAGS(12) := (BUFFER_SEMAPHORE = 0);
   end GET_HK_DIAG;


   procedure CONSTRUCTOR is
   begin
      EVENT_SIZE := 0;
      READING_INDEX := 0;
      WRITING_INDEX := 0;
      AMOUNT_OF_EVENTS := 0;
      BUFFER_LOCKED := TRUE;
      BUFFER_SEMAPHORE := 0;
   end CONSTRUCTOR;

end IASW_BUFFER;
