with SYSTEM;
with CSSW_COMMON_TYPES;
with IASW_COMMON;

package IASW_BUFFER is

   package CCT renames CSSW_COMMON_TYPES;

   EX_INDEX_OUT_OF_RANGE,
   EX_BUFFER_FULL,
   EX_BUFFER_EMPTY,
   EX_BUFFER_LOCKED
      : exception;

   subtype T_BUFFER_INDEX is CCT.T_UNSIGNED_16;

   --------------------------------------------------------------------
   -- Local types and data, made global because of possible patching --
   --------------------------------------------------------------------

   -- The amount of IASW_SUBBUFFER_n packages.
   -- WARNING! If this number is to be changed, it is not sufficient
   -- to change the number below.  See procedures: PREPARE_BUFFER and
   -- ACCES_EVENT.  Changes in files: Makefile and jemx.lnk are also
   -- necessary.
   K_NBUF : constant INTEGER := 10;

   -- Current size of a single data event.
   EVENT_SIZE : IASW_COMMON.T_EVENT_SIZE;

   -- The sizes of sub-buffers.
   SUBBUFFER_SIZES : array(1..K_NBUF) of T_BUFFER_INDEX;
   TOTAL_BUFFER_SIZE : T_BUFFER_INDEX;

   -- Current indices of the beginning and end of buffers.
   READING_INDEX : T_BUFFER_INDEX;
   WRITING_INDEX : T_BUFFER_INDEX;

   -- The amount of events stored in buffer.
   AMOUNT_OF_EVENTS : T_BUFFER_INDEX;

   -- The possibility of appending new events to the buffer.
   BUFFER_LOCKED : BOOLEAN;

   -- Semaphore controlling access to the buffer for two tasks
   --this will be used with IASW_COMMON.CAN_ENTER_CRITSEC(BUFFER_SEMAPHORE'address);
   BUFFER_SEMAPHORE : CCT.T_UNSIGNED_16;

   ----------------------------------------------------------------
   -- Local procedures, made global because of possible patching --
   ----------------------------------------------------------------

   procedure PREPARE_BUFFER;

   procedure ACCESS_EVENT(
      EVENT : in out IASW_COMMON.T_EVENT;
      INDEX : in T_BUFFER_INDEX;
      STORING : in BOOLEAN);

   -----------------------
   -- Global procedures --
   -----------------------

   procedure SET_EVENT_SIZE (EVENT_SIZE : in IASW_COMMON.T_EVENT_SIZE);
   procedure CLEAR_BUFFER;
   procedure PUT_EVENT (EVENT : in out IASW_COMMON.T_EVENT);
   procedure GET_EVENT (EVENT : in out IASW_COMMON.T_EVENT);
   procedure PEEK_EVENT(
      INDEX : in T_BUFFER_INDEX;
      EVENT : in out IASW_COMMON.T_EVENT);
   procedure DISCARD_EVENTS (AMOUNT : in T_BUFFER_INDEX);
   function IS_BUFFER_FULL return BOOLEAN;
   function IS_BUFFER_EMPTY return BOOLEAN;
   function GET_AMOUNT_OF_EVENTS return T_BUFFER_INDEX;
   function GET_TOTAL_BUFFER_SIZE return T_BUFFER_INDEX;
   procedure ENABLE_INPUT;
   procedure DISABLE_INPUT;
   function IS_BUFFER_LOCKED return BOOLEAN;

   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE);

   procedure GET_HK_DIAG( FLAGS : out IASW_COMMON.T_16_BITS );

   procedure CONSTRUCTOR;

end IASW_BUFFER;
