----------------------------------------------------------------------------
-- Module Name:        IASW_COMMON - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    Provides commonly used procedures
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with UNCHECKED_CONVERSION;
with CSSW_COMMON_TYPES;  use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;      use CSSW_IF_TYPES;
with DPE_CSSW;
with IASW_MAIN;
with IASW_MRTU;
with IASW_DFEE;
with IASW_HSL;
with IASW_MODES;
with IASW_STATE;
with IASW_SC_OUT;
with IASW_BUFFER;

package body IASW_COMMON is

   procedure I_AM_ALIVE( TASK_NO : in T_TASK_NO ) is
   begin
      if TASK_ALIVE_CNT(TASK_NO) < 255 then
         TASK_ALIVE_CNT(TASK_NO) := TASK_ALIVE_CNT(TASK_NO) + 1;
      else
         TASK_ALIVE_CNT(TASK_NO) := 0;
      end if;
   end I_AM_ALIVE;


   procedure START_TASK( TASK_ID : in CIT.T_IASW_TASK_ID ) is
   begin
      if DPE_CSSW.TASK_STATE(TASK_ID) = CIT.NON_OPERATIONAL then
         DPE_CSSW.START_TASK(TASK_ID);
      end if;
   end START_TASK;


   -- Z := X - Y - (C); C := borrow
   procedure SUBB(
      X,Y : in     CCT.T_UNSIGNED_16;
      Z   :    out CCT.T_UNSIGNED_16;
      C   : in out BOOLEAN
   ) is
   begin
      if C then
	 if X > Y then
	    Z := (X - Y) - 1;
	    C := FALSE;
	 else
	    Z := 16#FFFF# - (Y - X);
	 end if;
      else
	 if X >= Y then
	    Z := X - Y;
	 else
	    Z := (16#FFFF# - (Y - X)) + 1;
	    C := TRUE;
	 end if;
      end if;
   end SUBB;


   -- Z := X + Y + (C); C := carry;
   procedure ADDC(
      X, Y : in     CCT.T_UNSIGNED_16;
      Z    :    out CCT.T_UNSIGNED_16;
      C    : in out BOOLEAN
   ) is
   begin
      if C then
	 if (16#FFFF# - X) > Y then
	    Z := X + Y + 1;
	    C := FALSE;
	 else
	    Z := Y - (16#FFFF# - X);
	 end if;
      else
	 if (16#FFFF# - X) >= Y then
	    Z := X + Y;
	 else
	    Z := (Y - (16#FFFF# - X)) - 1;
	    C := TRUE;
	 end if;
      end if;
   end ADDC;


   procedure INC(
      X            : in out CCT.T_UNSIGNED_16;
      INCREMENT    : in CCT.T_UNSIGNED_16 := 1;
      X_AFTER_FFFF : in CCT.T_UNSIGNED_16 := 16#FFFF#
   ) is
   begin
      if X <= (16#FFFF#-INCREMENT) then
         X := X + INCREMENT;
      else
         X := X_AFTER_FFFF;
      end if;
   end INC;


   procedure INC32( X : in out CIT.T_U32_BITS ) is
   begin
      if X.LOW = 16#FFFF# then
         X.LOW := 0;
         if X.HIGH = 16#FFFF# then
            X.HIGH := 0;
         else
            X.HIGH := X.HIGH + 1;
         end if;
      else
         X.LOW := X.LOW + 1;
      end if;
   end INC32;


   function INC_MODULO_FFFF( X : in CCT.T_UNSIGNED_16 ) return CCT.T_UNSIGNED_16 is
   begin
      if X = 16#FFFF# then
         return 0;
      else
         return X + 1;
      end if;
   end INC_MODULO_FFFF;


   function INT2LONG( X : in INTEGER ) return LONG_INTEGER is
      type T_II is array(0..1) of INTEGER;

      II : T_II;

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => T_II, TARGET => LONG_INTEGER);
   begin
      II(1) := X;
      if X < 0 then
         II(0) := -1;
      else
         II(0) := 0;
      end if;
      return CVT(II);  --L;
   end INT2LONG;


   procedure GET_HK_DIAG(
      FLAGS : out IASW_COMMON.T_16_BITS;
      DIAG  : out T_DIAGNOSTIC_HK
   ) is
   begin
      DIAG := TASK_ALIVE_CNT;
   end GET_HK_DIAG;


   function GET_DUMMY_EVENT_TYPE (EVENT : in T_EVENT) return T_DUMMY_EVENT is
   begin
      if EVENT(2) /= 16#FFFF# then
         return NORMAL_EVENT;
      elsif EVENT(3) = 1 then
         return END_OF_DATA_MARKER;
      elsif EVENT(3) = 2 then
         return GREY_FILTER_REPORT;
      elsif EVENT(3) = 3 then
         return CALIBRATION_AMPLITUDE_REPORT;
      elsif EVENT(3) = 11 then
         return CHANGE_FORMAT_REQUEST;
      else
         return UNKNOWN_DUMMY_EVENT;
      end if;
   end GET_DUMMY_EVENT_TYPE;


   procedure CONSTRUCTOR is
   begin
      for i in T_TASK_NO loop
         TASK_ALIVE_CNT(i) := 0;
      end loop;
   end CONSTRUCTOR;

end IASW_COMMON;
