----------------------------------------------------------------------------
-- Module Name:        IASW_MODES - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    Procedures starting and stopping data taking.
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with CSSW_COMMON_TYPES; use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;
with DPE_CSSW;
with IASW_COMMON;
with IASW_SC_OUT;
with IASW_TM;
with IASW_HSL;
with IASW_BUFFER;
with IASW_DFEE;

package body IASW_MODES is

   function IS_SCF_VALID(
      SCF   : in T_SCIENCE_FORMAT;
      STATE : in IASW_DFEE.T_DFEE_STATE;
      SYNC  : in BOOLEAN
   ) return BOOLEAN is
   begin
      if not SYNC then
         return (SCF = K_SCF_TEST);
      end if;

      case STATE is
   when IASW_DFEE.K_STATE_DATA_TAKING =>
         return (SCF = K_SCF_TEST) or
                (SCF = K_SCF_FULL) or
                (SCF = K_SCF_SPECTRUM) or
                (SCF = K_SCF_RESTRICTED) or
                (SCF = K_SCF_SPECT_TIM) or
                (SCF = K_SCF_TIMING);

   when IASW_DFEE.K_STATE_CALIBRATION =>
         return (SCF = K_SCF_TEST) or
                (SCF = K_SCF_CALIBRATION);

   when IASW_DFEE.K_STATE_DIAGNOSTIC =>
         return (SCF = K_SCF_TEST) or
                (SCF = K_SCF_DIAGNOSTIC);

   when others =>
         return FALSE;

      end case;
   end IS_SCF_VALID;


   function GET_DEFAULT_SCF(
      STATE : in IASW_DFEE.T_DFEE_STATE;
      SYNC  : in BOOLEAN
   ) return T_SCIENCE_FORMAT is
   begin
      if not SYNC then
         return K_SCF_TEST;
      end if;
      case STATE is
   when IASW_DFEE.K_STATE_DATA_TAKING  => return K_SCF_FULL;
   when IASW_DFEE.K_STATE_CALIBRATION  => return K_SCF_CALIBRATION;
   when IASW_DFEE.K_STATE_DIAGNOSTIC   => return K_SCF_DIAGNOSTIC;
   when others                         => return K_SCF_UNDEFINED;
      end case;
   end GET_DEFAULT_SCF;


   function CALCULATE_EVENT_SIZE(   -- local
      STATE : in IASW_DFEE.T_DFEE_STATE
   ) return IASW_COMMON.T_EVENT_SIZE is
   begin
      case STATE is
   when IASW_DFEE.K_STATE_DATA_TAKING => return 5;
   when IASW_DFEE.K_STATE_CALIBRATION => return 36;
   when IASW_DFEE.K_STATE_DIAGNOSTIC  => return 42;
   when others                        => return 0;
      end case;
   end CALCULATE_EVENT_SIZE;


   procedure STOP_DATA_TAKING (IMM : in BOOLEAN := FALSE) is
   begin
      if IMM then
         IASW_HSL.HARD_STOP;
         IASW_SC_OUT.SCTM_STOP(IASW_SC_OUT.K_SCTM_HARD_STOP);
         loop
            exit when (IASW_HSL.STOPPED and IASW_SC_OUT.STOPPED);
            DPE_CSSW.DELAY_RELATIVE(1);
         end loop;
      else
         IASW_SC_OUT.SCTM_STOP(IASW_SC_OUT.K_SCTM_DELAYED_SOFT_STOP);
      end if;
   end STOP_DATA_TAKING;


   procedure START_DATA_TAKING (
      STATE : in IASW_DFEE.T_DFEE_STATE;
      SYNC  : in BOOLEAN;
      SF1   : in T_SCIENCE_FORMAT;
      SF2   : in T_SCIENCE_FORMAT;
      GFL1  : in CCT.T_UNSIGNED_16;
      GFL2  : in CCT.T_UNSIGNED_16
   ) is
   begin
      STOP_DATA_TAKING (IMM => TRUE);

      IASW_HSL.START_DATA_TAKING(
         CALCULATE_EVENT_SIZE (STATE),
         SF1, SF2, GFL1, GFL2, SYNC );

      IASW_SC_OUT.START_DATA_TAKING;
   end START_DATA_TAKING;


   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE) is
   begin
      null;
   end GET_DPE_STATE;


   procedure CONSTRUCTOR is
   begin
      null;
   end CONSTRUCTOR;

end IASW_MODES;
