----------------------------------------------------------------------------
-- Module Name:        IASW_SC_OUT - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with SYSTEM;             use SYSTEM;
with CSSW_COMMON_TYPES;  use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;      use CSSW_IF_TYPES;
with DPE_CSSW;
with UNCHECKED_CONVERSION;
with IASW_COMMON;
with IASW_TM;
with IASW_BUFFER;
with IASW_EVSORT;
with IASW_MODES;
with IASW_DFEE;     --HV settings, Energy table
with IASW_TIME;
with IASW_GF;

package body IASW_SC_OUT is

   procedure ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(GF : in CCT.T_UNSIGNED_16) is
   begin
      if GF in IASW_GF.T_GREY_FILTER_INDICATOR then
	 GREY_FILTER_REPORTED := GF;
      else
	 GREY_FILTER_REPORTED := IASW_GF.GREY_FILTER_100;
      end if;
   end ACKNOWLEDGE_OF_GREY_FILTER_CHANGE;


   function NO_DATA_IN_THE_BUFFER return BOOLEAN is
   begin
      if IASW_BUFFER.IS_BUFFER_EMPTY then
	 if (COMPRESSION_PROC_PHASE = CPP_SOFT_STOP) and
            (B1.STATE = OBS_BUFFER_PARTIALLY_FILLED)
	 then
	    B1.STATE := OBS_SEQUENCE_COMPLETED;
	 end if;
	 return TRUE;
      else
         return FALSE;
      end if;
   end NO_DATA_IN_THE_BUFFER;


   procedure SET_SC_TM_DATA_HEADER(
   --
   -- procedure dealing with the new format of SC TM Data Header.
   -- TYP, STYP and OBT will be fitt into the field used previously
   -- to store only OBT.
   --
      TM_BUFF  : out CIT.T_SCIENCE_DATA_FIELD;
      TYP      : in IASW_COMMON.T_PACKET_TYPE;
      STYP     : in IASW_COMMON.T_PACKET_SUBTYPE;
      OBT      : in CIT.T_OBT
   ) is
      type T_SC_DATA_HDR is record
         T : IASW_COMMON.T_PACKET_TYPE;
         S : IASW_COMMON.T_PACKET_SUBTYPE;
         O1,O2,O3,O4,O5 : CCT.T_UNSIGNED_8;
      end record;
      for T_SC_DATA_HDR use record
         T  at 0 range 0..3;
         S  at 0 range 4..7;
         O1 at 0 range 8..15;
         O2 at 1 range 0..7;
         O3 at 1 range 8..15;
         O4 at 2 range 0..7;
         O5 at 2 range 8..15;
      end record;

      HDR : T_SC_DATA_HDR;
         for HDR use at TM_BUFF'address;
pragma If(((HDR'size+15)/16) > TM_BUFF'length);
   error: Violated safety of 'use at'
pragma End if;

      O : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(1..6);
         for O use at OBT'address;
pragma If(O'length*8 > OBT'size);
   error: Violated safety of 'use at'
pragma End if;

   begin
      HDR.T := TYP;
      HDR.S := STYP;
      HDR.O1 := O(1);
      HDR.O2 := O(2);
      HDR.O3 := O(3);
      HDR.O4 := O(4);
      HDR.O5 := O(5);
   end SET_SC_TM_DATA_HEADER;


   procedure SC_TM_OUT (SC_BUF : in out T_OUT_BUF) is
   --
   -- Sends a given TM packet.
   --
      TM_DATA : CIT.T_SCIENCE_DATA_FIELD;
      TM_DONE : BOOLEAN;
   begin
      if SC_BUF.STATE < OBS_BUFFER_COMPLETED then
	 return;
      end if;

      TM_DATA := SC_BUF.DATA;   -- PUT_TM() wants to have this on stack.
      IASW_TM.PUT_TM(SC_BUF.APID, TM_DATA, TM_DONE);

      if TM_DONE then
	 if (SC_BUF.STATE = OBS_BUFFER_COMPLETED) then
	    -- used only in Restricted Im. F.
	    SC_BUF.STATE := OBS_BUFFER_NEED_INIT;
	 else
	    SC_BUF.STATE := OBS_SEQUENCE_NEED_INIT;
	 end if;
      else
         -- PUT_TM() was unsuccessfull.
         DPE_CSSW.DELAY_RELATIVE(1);
      end if;
   end SC_TM_OUT;


   procedure PUT_BYTE_TO_TM_DATA(
   --
   -- this is used localy inside compressing procedures.
   -- it has been separated because of an error in compiler - space for
   -- array DATA was allocated on stack even when using 'for DATA use at ...'
   --
      BYTE_DATA  : in CCT.T_UNSIGNED_8;
      DATA_ADDR  : in SYSTEM.ADDRESS;
      DATA_INDEX : in out INTEGER
   ) is
      DATA : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(1..440);
         for DATA use at DATA_ADDR;
         -- this 'use at' can be not safe if the parameter DATA_ADDR points
         -- to wrong place in memory, or if the displacement - DATA_INDEX is
         -- too big.
   begin
      DATA_INDEX := DATA_INDEX + 1;
      DATA(DATA_INDEX) := BYTE_DATA;
   end PUT_BYTE_TO_TM_DATA;


   function GET_DELTA_TIME(
   --
   -- Calculates difference of time between two adjacent data events.
   -- This procedure in a natural way works good with CC2 rolled over FFFF,
   --
      CC2, TC2, CC1, TC1 : in CCT.T_UNSIGNED_16
   ) return CIT.T_U32_BITS is
      C : BOOLEAN;
      Z : CIT.T_U32_BITS;
   begin
      C := FALSE;
      IASW_COMMON.SUBB(TC2, TC1, Z.LOW, C);
      IASW_COMMON.SUBB(CC2, CC1, Z.HIGH, C);
      return Z;
   end GET_DELTA_TIME;


   function TRUNC16(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16 is
   begin
      if X.HIGH /= 0 then
	 return 16#FFFF#;
      else
	 return X.LOW;
      end if;
   end TRUNC16;


   function MUL8(X : in CIT.T_U32_BITS) return CIT.T_U32_BITS is
      Z : CIT.T_U32_BITS;
   begin
      if X.HIGH < 16#2000# then
         Z.HIGH := X.HIGH*8 + (X.LOW / 16#2000#);
         Z.LOW := (X.LOW mod 16#2000#)*8;
      else
         Z.HIGH := 16#FFFF#;
         Z.LOW := 16#FFFF#;
      end if;
      return Z;
   end MUL8;


   function DIV4(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16 is
   begin
      return (X.LOW / 4) + (X.HIGH mod 4) * 16#4000#;
   end DIV4;


   function DIV1024(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16 is
   begin
      return (X.LOW / 1024) + (X.HIGH mod 1024) * 64;
   end DIV1024;


   function CLEAR10BITS(X : in CCT.T_UNSIGNED_16) return CCT.T_UNSIGNED_16 is
   begin
      return (X / 1024) * 1024;
   end CLEAR10BITS;


   function COORD_8BIT(
   --
   -- Recalculates scale of X/Y coordinates.  Input scale: 0..1023,
   -- output scale: 1..255.
   --
      X : in CCT.T_UNSIGNED_16;
      EVENT_POSITION_TABLE : in T_EVENT_POSITION_TABLE
   ) return CCT.T_UNSIGNED_8 is
   begin
      return EVENT_POSITION_TABLE(IASW_COMMON.CVT(X mod 1024));
   end COORD_8BIT;


   --------------------------------------------------------------------
   -- DFEE's bugg
   --

   procedure FIXBUGG1_INIT is
   -- this is called at the beginning of data taking
   begin
      FIXBUGG1.VALID := FALSE;
   end FIXBUGG1_INIT;

   procedure FIXBUGG1_DO(EVENT : in out IASW_COMMON.T_EVENT) is
   -- this is called after each GET_EVENT or PEEK_EVENT.
   begin
      if (EVENT(1) = 0) and FIXBUGG1.VALID and (EVENT(0) = FIXBUGG1.CC) then
         -- fix the bugg - do what DFEE didn't do
         IASW_COMMON.INC(X => EVENT(0), INCREMENT => 1, X_AFTER_FFFF => 0);
      end if;
      FIXBUGG1.PREV_CC := FIXBUGG1.CC;
      FIXBUGG1.CC := EVENT(0);
      FIXBUGG1.VALID := TRUE;
   end FIXBUGG1_DO;

   procedure FIXBUGG1_BACK is
   -- this is called when after PEEK_EVENT there is no DISCARD_EVENT, then
   -- the same event will be PEEK-ed next time.
   begin
      FIXBUGG1.CC := FIXBUGG1.PREV_CC;
   end FIXBUGG1_BACK;


   --------------------------------------------------------------------
   -- Count Rate format procedures: CR_INIT, CR_OUT, CR_EVENT
   --

   procedure CR_INIT is
   --
   -- Sets all CR variables and buffers to their initial values.  Usually
   -- It will be called only once, at the beginning of new data mode or
   -- compression format.  It will also be used in case of discontinuity
   -- in time of events.
   --
   begin
      for I in CR_DATA'Range loop
	 CR_DATA(I) := 0;
      end loop;
      CR_INDEX     := -1;   -- this means empty CR_DATA
      CR_GF_INDEX  :=  0;
      B2.STATE := OBS_SEQUENCE_NEED_INIT;
      B2.APID  := IASW_COMMON.APID_TM_SC_COUNT_RATE;
   end CR_INIT;


   procedure CR_OUT is
   --
   -- Prepares and sends TM packet containing CR data.  The data, which
   -- don't fitt into TM packet are moved to the beginning of the CR
   -- buffer and will be sent next time.
   --
      type T_GFCH is array(1..8) of T_GF_AND_CHANNEL;
      type T_HEADER is record
	 OBT    : CIT.T_OBT;            -- OBT of the first event.
	 GFCH   : T_GFCH;               -- Grey filter specifiers.
	 TCHNUM : INTEGER;              -- Number of time channels in packet.
	 COUNT0 : T_CR_DATA;            -- CR for the first time channel.
      end record;
      type T_10_6 is record
	 N_10 : CCT.T_UNSIGNED_10;
	 N_6  : CCT.T_UNSIGNED_6;
      end record;
      for T_10_6 use record
	 N_10 at 0 range 0..9;
	 N_6  at 0 range 10..15;
      end record;
      -- type T_10_6 is used in the calculation of new time-base
      -- CC0,CT0 when moving N channels forward:
      -- new (CC0,CT0) := old (CC0,CT0) + N*1024;

      HDR : T_HEADER;
         for HDR use at B2.DATA'Address;
pragma If(((HDR'size+15)/16) > B2.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      HDRLEN : constant INTEGER := (HDR'Size / 16);
      DATA_ADDR : SYSTEM.ADDRESS := B2.DATA(HDRLEN+1)'Address;
      PC, AC : T_CR_DATA;
      J, DIFF, N, N_GF : INTEGER;
      WW : CCT.T_UNSIGNED_16;
      BB : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..1);
      N_10_6 : T_10_6;
         for DIFF use at BB'Address;
         for WW use at BB'Address;
         for N_10_6 use at N'Address;
         -- all those 'use at' are safe.
      C   : BOOLEAN;
      GFI : IASW_GF.T_GREY_FILTER_INDICATOR;
      DIFF_OK : BOOLEAN;
   begin
      -- Check if there is any data to send.
      -- Dont consider the last time channel, because it is most probably
      -- not yet completed.  That's why "<=", not "<".
      if CR_INDEX <= 0 then
	 return; -- There is no data in the CR buffer.
      end if;

      -- Send OEM when we have to drop unsent TM packet.
      -- It might happen, that the previous TM packet couldn't be sent
      -- and now we can't wait - it must be dropped.
      if B2.STATE >= OBS_BUFFER_COMPLETED then
	 IASW_TM.PUT_OEM (IASW_TM.EV_TM_LOST, INTEGER(B2.APID));
      end if;

      -- Prepare header of TM packet.
      IASW_TM.INIT_SCIENCE_TM (B2.DATA);
      SET_SC_TM_DATA_HEADER(
         B2.DATA, 3,8,
         IASW_TIME.GET_OBT_OF_EVENT(CR_CC0, CLEAR10BITS(CR_TC0)));
      HDR.COUNT0  := CR_DATA(0);  -- first CR channel goes into the header

      -- Pack as many CR channels as possible.
      PC := HDR.COUNT0;
      B2.INDEX := 0;   -- index of bytes array
      N  := 1;         -- counter of time channels moved to TM packet
      for I in 1..(CR_INDEX - 1) loop
      -- Dont consider the last time channel.  That's why (CR_INDEX-1).
	 AC := CR_DATA(I);
	 begin
	    DIFF := INTEGER(AC) - INTEGER(PC);
	    DIFF_OK := TRUE;
	 exception
	    when others =>
	       DIFF_OK := FALSE;
	 end;
	 if (not DIFF_OK) or (DIFF > 126) or (DIFF < -127) then
	    exit when (B2.INDEX + 3) > (CR_TM_CHANNELS - 1);
            -- array of bytes is (CR_TM_CHANNELS-1) long
	    WW := AC;
	    PUT_BYTE_TO_TM_DATA( 127, DATA_ADDR, B2.INDEX);
	    PUT_BYTE_TO_TM_DATA( BB(0), DATA_ADDR, B2.INDEX);
	    PUT_BYTE_TO_TM_DATA( BB(1), DATA_ADDR, B2.INDEX);
	 else
	    exit when (B2.INDEX + 1) > (CR_TM_CHANNELS - 1);
	    PUT_BYTE_TO_TM_DATA( BB(1), DATA_ADDR, B2.INDEX);
	 end if;
	 PC   := AC;
	 N    := N + 1;
      end loop;

      HDR.TCHNUM := N;  -- number of time channels copied to TM packet

      -- copy the list of 'GF/index', but only those, where 'index' points
      -- to the time channels, which are present in TM packet.
      N_GF := 0;    -- counter of 'GF/index' copied to TM packet
      GFI  := IASW_GF.GREY_FILTER_100;
      for I in 1..CR_GF_INDEX loop
	 exit when INTEGER(CR_GF_DATA(I).CRI) >= N;
	 exit when I not in HDR.GFCH'Range;
	 HDR.GFCH(I) := CR_GF_DATA(I);
	 N_GF := I;   -- number of GFs copied to TM packet
	 GFI  := IASW_GF.T_GREY_FILTER_INDICATOR(CR_GF_DATA(I).GFI);
      end loop;

      -- Update CR variables.  Unused time channels and GF/index must be
      -- shifted to the beginning of the buffer.  Indices must me modified,
      -- because of shift of time channels.
      -- Time reference (CR_CC0,CR_TC0) must also be modified.

      for I in 0..(CR_INDEX - N) loop
	 CR_DATA(I) := CR_DATA(I + N);
      end loop;
      for I in (CR_INDEX - N + 1)..CR_INDEX loop
	 CR_DATA(I) := 0;
      end loop;
      CR_INDEX := CR_INDEX - N;

      -- at least one GF must be ready for the next packet, set it to
      -- the last used GF.
      J := 1;
      CR_GF_DATA(J) := (
         GFI => CCT.T_UNSIGNED_5(GFI),
         CRI => 0
      );
      -- move pending GFs to the beginning of the buffer
      for I in (N_GF + 1)..CR_GF_INDEX loop
	 if INTEGER(CR_GF_DATA(I).CRI) >= N then
            -- suppress repeated GFs.
            if CR_GF_DATA(I).GFI /= CR_GF_DATA(J).GFI then
	       J := J + 1;
	       CR_GF_DATA(J) := (
                  GFI => CR_GF_DATA(I).GFI,
	          CRI => CR_GF_DATA(I).CRI - CCT.T_UNSIGNED_11(N)
               );
            end if;
         else
            -- if for some reason the index is too small, the GF/index
            -- entry will be ignored.
            null;
	 end if;
      end loop;
      CR_GF_INDEX := J;

      -- new (CC0,CT0) := old (CC0,CT0) + N*1024;
      -- N_10_6 is overlayed over N.
      C := FALSE;
      IASW_COMMON.ADDC (CR_TC0, 1024 * CCT.T_UNSIGNED_16(N_10_6.N_6), CR_TC0, C);
      IASW_COMMON.ADDC (CR_CC0, CCT.T_UNSIGNED_16(N_10_6.N_10),       CR_CC0, C);

      -- Signal the fact that the buffer is ready to be sent.
      B2.STATE := OBS_SEQUENCE_COMPLETED;

      -- Try to send prepared TM packet.
      SC_TM_OUT (B2);
   end CR_OUT;


   procedure CR_EVENT(
   --
   -- Incremets count-rate data counters and sends TM packet when
   -- necessary.
   --
      CC, TC : in CCT.T_UNSIGNED_16;
      DET    : in IASW_COMMON.T_DUMMY_EVENT;
      GFI    : in IASW_GF.T_GREY_FILTER_INDICATOR
   ) is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      INDEX : INTEGER;
      RESTART : BOOLEAN;
      W : CCT.T_UNSIGNED_16;
   begin
      -- Send TM pending data, if previous try was unsuccessfull.
      if B2.STATE >= OBS_BUFFER_COMPLETED then
	 SC_TM_OUT (B2);
      end if;

      RESTART := TRUE;   -- preliminary

      -- Calculate CR index.  Initialize CR variables on the first event.
      if CR_INDEX >= 0 then    -- this is continuation of the block
	 W := DIV1024(GET_DELTA_TIME(CC, TC,  CR_CC0, CR_TC0));
         if W < 16#8000# then
            INDEX := INTEGER(W);
            RESTART := (INDEX not in CR_DATA'range);
         end if;
      elsif DET /= IASW_COMMON.NORMAL_EVENT then
	 return; -- Ignore GF reports before first normal event.
      end if;

      if RESTART then
      -- this may happen at the first event in a seesion, or after a long
      -- time gap without any events.
	 CR_OUT;  -- first, empty the buffer with previously collected data.
	 CR_INIT; -- resume session with a clear buffer.
         -- Initiate the new block. The time of the first event, truncated
         -- to the width of time channel will be a reference for the block.
	 CR_CC0 := CC;
	 CR_TC0 := (TC / 1024) * 1024;
	 INDEX := 0;
	 CR_GF_INDEX := 1;
	 CR_GF_DATA(CR_GF_INDEX) := (
	    GFI => CCT.T_UNSIGNED_5(GFI),
	    CRI => CCT.T_UNSIGNED_11(INDEX)
         );
      end if;

      if DET = IASW_COMMON.NORMAL_EVENT then
	 -- Update counter in the corresponding time channel
	 IASW_COMMON.INC (CR_DATA(INDEX));
      elsif CCT.T_UNSIGNED_5(GFI) /= CR_GF_DATA(CR_GF_INDEX).GFI then
         -- this is GF change report, GF reported is differente
         -- than the previous one.
	 -- Append GF data.
	 if (CR_GF_INDEX + 1) in CR_GF_DATA'Range then
            -- this should always be true, because
            -- CR_GF_DATA'Range > CR_TM_GF_CHANNELS, and if only
            -- CR_GF_INDEX > CR_TM_GF_CHANNELS, then the block will be closed,
            -- (see below).
	    CR_GF_INDEX := CR_GF_INDEX + 1;
	    CR_GF_DATA(CR_GF_INDEX) := (
	       GFI => CCT.T_UNSIGNED_5(GFI),
	       CRI => CCT.T_UNSIGNED_11(INDEX)
            );
	 end if;
      end if;

      -- update the maximal index
      if INDEX > CR_INDEX then
	 CR_INDEX := INDEX;
      end if;

      -- Send CR data to TM when we are out of free buffer space.
      if (CR_INDEX > CR_TM_CHANNELS) or      -- main buffer full
         (CR_GF_INDEX > CR_TM_GF_CHANNELS)   -- too many GF changes
      then
	 CR_OUT;
      end if;
   end CR_EVENT;


   procedure REQUEST_FOR_SCF_SWITCHING( RQ : in CCT.T_UNSIGNED_16 ) is
   begin
      SCF_WANTED := RQ mod 256;
      EVSIZE_WANTED := RQ / 256;
      SCF_SWITCH_REQUEST := TRUE;
   end REQUEST_FOR_SCF_SWITCHING;


   procedure SCTM_STOP(STOP_TYPE : in T_SCTM_STOP_TYPE) is
   begin
      if STOP_TYPE > SCTM_STOP_TYPE then
         SCTM_STOP_TYPE := STOP_TYPE;
         SOFT_STOP_DELAY := 0;
      end if;
   end SCTM_STOP;


   procedure SCF_UNDEFINED is
   -- the only job for this is to react on dummy events
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      EVENT : IASW_COMMON.T_EVENT; -- Event currently processed.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
   begin
      loop
         exit when IASW_BUFFER.IS_BUFFER_EMPTY;
	 IASW_BUFFER.GET_EVENT (EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

         if DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

            ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- no more events expected, exit the loop
            exit;

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

            REQUEST_FOR_SCF_SWITCHING(EVENT(4));
            -- exit the loop to give chance to execute this request in TASK_SC_OUT.
            exit;

	 else
	    null;  -- another event-record, skip it.
	 end if;
      end loop;
   end SCF_UNDEFINED;


   procedure SCF_TEST is
      EVENT : IASW_COMMON.T_EVENT;      -- Event currently processed.
      JI    : INTEGER;
      JW    : CCT.T_UNSIGNED_16;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
         SET_SC_TM_DATA_HEADER(B1.DATA, 0,7, DPE_CSSW.GET_CURRENT_OBT);
	 B1.INDEX := 3;   -- length of Data Header
	 TM_EVENTS_PROCESSED := 0;
	 B1.STATE := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);

	 if B1.STATE = OBS_BUFFER_EMPTY then
	    B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;
	 end if;

	 for I in 0 .. EVENT_SIZE-1 loop
	    JW := I;
	    JI := B1.INDEX + IASW_COMMON.CVT(JW) + 1;
	    B1.DATA (JI) := IASW_COMMON.CVT(EVENT (I));
	 end loop;

	 B1.INDEX := B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE);
	 TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;

	 if ((B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE)) > K_MAX_TM_DATA_INDEX) then
	    B1.STATE := OBS_SEQUENCE_COMPLETED;
	 end if;

      end loop;
   end SCF_TEST;


   procedure SCF_FULL is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT : CIT.T_OBT;           -- OBT of the first event in the package.
	 NEV : CCT.T_UNSIGNED_16;   -- number of events in the package
	 GF  : IASW_GF.T_GREY_FILTER_INDICATOR;  -- for the first event
      end record;
      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      type T_PACKED_WORD_1 is record
         X, Y : T_UNSIGNED_8;
      end record;

      for T_PACKED_WORD_1 use record
         X at 0 range 0 ..  7;
         Y at 0 range 8 .. 15;
      end record;

      type T_PACKED_WORD_2 is record
         ENERGY     : T_UNSIGNED_8;
         DELTA_TIME : T_UNSIGNED_8;
      end record;

      for T_PACKED_WORD_2 use record
         ENERGY     at 0 range 0 ..  7;
         DELTA_TIME at 0 range 8 .. 15;
      end record;

      function CVT is new UNCHECKED_CONVERSION
	(SOURCE => T_PACKED_WORD_1, TARGET => INTEGER);

      function CVT is new UNCHECKED_CONVERSION
	(SOURCE => T_PACKED_WORD_2, TARGET => INTEGER);

      EVENT            : IASW_COMMON.T_EVENT; -- Event currently processed.
      DELTA_TIME       : CCT.T_UNSIGNED_16;   -- Time between events.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
      APPENDED         : BOOLEAN;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	 B1.INDEX            := HDR'size / 16;
	 TM_EVENTS_PROCESSED := 0;
	 HDR.NEV             := TM_EVENTS_PROCESSED;
	 B1.STATE            := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);
         FIXBUGG1_DO(EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

         if DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

            REQUEST_FOR_SCF_SWITCHING(EVENT(4));
            -- this will become effective only after completting the packet

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- no more events expected, exit the loop
            exit;

         else   -- normal event, or GF report

	    if B1.STATE = OBS_BUFFER_EMPTY then
	       -- Fill in the packet header for the first event.
               SET_SC_TM_DATA_HEADER( B1.DATA, 0,8,
                  IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), EVENT(1)));

	       HDR.GF := GREY_FILTER_REPORTED;

               if THIS_IS_START_OF_FORMAT then
                  -- without this the GET_DELTA_TIME(), which is near
                  -- below would have no PREV_xx for reference.
                  -- this is done only once at the beginning, so next time
                  -- the first event in the packet will have delta time
                  -- relative to the previous event (last one in the
                  -- previous packet), not zero.
                  PREV_CC := EVENT(0);
                  PREV_TC := EVENT(1);
                  THIS_IS_START_OF_FORMAT := FALSE;
               end if;
	    end if;

         end if;

	 APPENDED := FALSE;   -- preliminary

	 if DUMMY_EVENT_TYPE = IASW_COMMON.NORMAL_EVENT then

            -- delta time relative to the time of previous real event
	    DELTA_TIME := TRUNC16(GET_DELTA_TIME(
               EVENT(0), EVENT(1),
               PREV_CC, PREV_TC));

	    -- Append event to the end of the packet.
	    B1.DATA(B1.INDEX + 1) := CVT(T_PACKED_WORD_1'(
               X => COORD_8BIT(EVENT(4),EVENT_X_TABLE),
               Y => COORD_8BIT(EVENT(3),EVENT_Y_TABLE)));

	    if DELTA_TIME >= 255 then
	       B1.DATA(B1.INDEX + 2) := CVT(T_PACKED_WORD_2'(
                  ENERGY     => CCT.T_UNSIGNED_8(
                     ENERGY_TABLE(IASW_COMMON.CVT(EVENT(2) mod 4096))),
                  DELTA_TIME => 255));
	       B1.DATA(B1.INDEX + 3) := IASW_COMMON.CVT(DELTA_TIME);
	       B1.INDEX := B1.INDEX + 3;
	    else
	       B1.DATA(B1.INDEX + 2) := CVT(T_PACKED_WORD_2'(
                  ENERGY     => CCT.T_UNSIGNED_8(
                     ENERGY_TABLE(IASW_COMMON.CVT(EVENT(2) mod 4096))),
                  DELTA_TIME => CCT.T_UNSIGNED_8(DELTA_TIME)));
	       B1.INDEX := B1.INDEX + 2;
	    end if;

	    -- Store HKCC and TC for further use.
            -- only real events time is stored.
	    PREV_CC := EVENT(0);
	    PREV_TC := EVENT(1);

	    APPENDED := TRUE;

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

            ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));
	    B1.DATA(B1.INDEX + 1) := 0;   -- X/Y = 0/0
	    B1.DATA(B1.INDEX + 2) := IASW_COMMON.CVT(GREY_FILTER_REPORTED);
	    B1.INDEX              := B1.INDEX + 2;

	    APPENDED := TRUE;

	 else
	    null;  -- another dummy event-record, skip it.
	 end if;

	 if APPENDED then
	    B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;

	    -- Increment the counter of events in the packet.
	    TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;
	    HDR.NEV             := TM_EVENTS_PROCESSED;

	    -- Check if this was last event in the packet.
	    if ((B1.INDEX + 3) > K_MAX_TM_DATA_INDEX) then
	       B1.STATE := OBS_SEQUENCE_COMPLETED;
	    end if;
	 end if;
      end loop;
   end SCF_FULL;


   procedure SCF_RESTRICTED is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_PACKED_EVENT is record
         DELTA_POSITION : CCT.T_UNSIGNED_5;
         ENERGY         : CCT.T_UNSIGNED_3;
      end record;
      for T_PACKED_EVENT use record
         DELTA_POSITION at 0 range  0 ..  4;
         ENERGY         at 0 range  5 ..  7;
      end record;

      type T_HEADER is record
	 OBT       : CIT.T_OBT;         -- OBT of the first event.
	 EVENTS    : CCT.T_UNSIGNED_16; -- Amount of events in the package.
         GF        : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(1..2);
	 DELTA_TIME: CIT.T_U32_BITS; -- Delta time of the last event.
         -- DELTA_TIME keeps seconds in HIGH and fractial os sec. in LOW
	 XY1_POS   : CCT.T_UNSIGNED_16;   -- X and Y pos for first event.
	 XY2_POS   : CCT.T_UNSIGNED_16;   -- X and Y pos for last event.
	 SEQ_NUM   : CCT.T_UNSIGNED_16; -- Sequence # for this packet.
      end record;

      HDRLEN : constant INTEGER := T_HEADER'size/16;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      DATA_ADDR : SYSTEM.ADDRESS := B1.DATA(HDRLEN + 1)'address;

      EVENT : IASW_COMMON.T_EVENT;        -- Event currently processed.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
      BB : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..1);
      PEV : T_PACKED_EVENT;
         for PEV use at BB'Address;
         -- this 'use at' is safe.
      W : CCT.T_UNSIGNED_16;
         for W use at BB'Address;
         -- this 'use at' is safe.
      DATA_INDEX : INTEGER;
      EVENT_FOUND : BOOLEAN;
      FIRST_EVENT_INDEX, LAST_EVENT_INDEX : IASW_BUFFER.T_BUFFER_INDEX;
      XY1 : CCT.T_UNSIGNED_16;
      DELTA_POS : CCT.T_UNSIGNED_16;
      CC1,TC1 : CCT.T_UNSIGNED_16;
      CRI : INTEGER;
      EVNO : IASW_BUFFER.T_BUFFER_INDEX;
   begin
      if B1.STATE = OBS_SEQUENCE_NEED_INIT then
      -- this means that process should wait for 2400 events in the buffer

	 if (IASW_BUFFER.GET_AMOUNT_OF_EVENTS < RIF_MAX_EVENTS) and
            (RIF_TIME < 5) and
            (COMPRESSION_PROC_PHASE /= CPP_SOFT_STOP)
         then
	    -- not yet enough events collected in the buffer.
	    return;
	 end if;

         -- start counting time for the next image
         RIF_TIME := 0;

	 EVNO := IASW_BUFFER.GET_AMOUNT_OF_EVENTS;
	 if EVNO < 1 then
            -- EVNO usually will be >= RIF_MAX_EVENTS, but can be
            -- smaller in two cases:
            --  1) CPP_SOFT_STOP,
            --  2) RIF_TIME >= 5
	    return;
	 end if;
	 if EVNO > RIF_MAX_EVENTS then
	    EVNO := RIF_MAX_EVENTS;
	 end if;

	 -- Step through all collected events, build the sorting keys list
	 -- and count-rate data.  Find the first and the last true event
	 -- (not dummy).  Find also grey filters corresponding to the first
	 -- and the last events.
         -- modifies GREY_FILTER_REPORTED while peaking events.

	 IASW_EVSORT.CLEAR_SORTING_KEYS_LIST;
	 EVENT_FOUND := FALSE;
         RIF_EVENTS_NO := 0;

	 for I in 0 .. EVNO - 1 loop

	    IASW_BUFFER.PEEK_EVENT (INDEX => I, EVENT => EVENT);
            FIXBUGG1_DO(EVENT);
	    DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

            -- count event-records considered.  At the end RIF_EVENTS_NO
            -- can be smaller than EVNO, because the 'for ... loop' can
            -- be broken before the end.
            RIF_EVENTS_NO := RIF_EVENTS_NO + 1;

	    if DUMMY_EVENT_TYPE = IASW_COMMON.NORMAL_EVENT then

	       CR_EVENT(
                  EVENT(0), EVENT(1),
                  IASW_COMMON.NORMAL_EVENT,
                  GREY_FILTER_REPORTED);
	       BB(0) := COORD_8BIT (EVENT(4),EVENT_X_TABLE);
	       BB(1) := COORD_8BIT (EVENT(3),EVENT_Y_TABLE);
	       IASW_EVSORT.ADD_SORTING_KEY (BUFFER_POS => I, KEY => W);
	       if not EVENT_FOUND then    -- first event
		  FIRST_EVENT_INDEX := I;
		  RIF_GF1 := GREY_FILTER_REPORTED;
		  EVENT_FOUND := TRUE;
	       end if;
	       LAST_EVENT_INDEX := I;
               RIF_GF2 := GREY_FILTER_REPORTED;

	    elsif DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

               ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));
	       CR_EVENT(
                  EVENT(0), EVENT(1),
                  IASW_COMMON.GREY_FILTER_REPORT,
                  GREY_FILTER_REPORTED);

            elsif DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

               REQUEST_FOR_SCF_SWITCHING(EVENT(4));
               -- this will become effective only after completting the image

	    elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

               SCTM_STOP(K_SCTM_SOFT_STOP);
               -- no more events expected, exit the loop
               exit;

	    else
	       null;
	    end if;
	 end loop;

	 if not EVENT_FOUND then
	    IASW_BUFFER.DISCARD_EVENTS (AMOUNT => RIF_EVENTS_NO);
	    return;
	 end if;

         -- collect some data for TM packets headers
	 IASW_BUFFER.PEEK_EVENT (INDEX => FIRST_EVENT_INDEX, EVENT => EVENT);
	 CC1 := EVENT(0);
	 TC1 := EVENT(1);
	 RIF_OBT := IASW_TIME.GET_OBT_OF_EVENT (CC1, TC1);

	 IASW_BUFFER.PEEK_EVENT (INDEX => LAST_EVENT_INDEX, EVENT => EVENT);
	 RIF_DELTA_TIME := MUL8(GET_DELTA_TIME (EVENT(0), EVENT(1), CC1, TC1));

	 -- Sort events before packing them.
	 IASW_EVSORT.SORT_EVENTS;
         RIF_SORTED_EVENTS_NO := IASW_EVSORT.GET_AMOUNT_OF_SORTED_EVENTS;
	 RIF_EVENTS_PACKED := 0;
	 RIF_PACKETS_SENT  := 0;
      end if;

      -- When the program comes to this point, it has already sorted some
      -- events and now can put them into telemetry packets.

      -- This place is practicaly the entry point of the procedure, when
      -- it is continuing with producing of TM.

      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
         SET_SC_TM_DATA_HEADER( B1.DATA, 1,8, RIF_OBT);
	 HDR.EVENTS     := 0;
	 HDR.GF(1)      := CCT.T_UNSIGNED_8(RIF_GF1);
	 HDR.GF(2)      := CCT.T_UNSIGNED_8(RIF_GF2);
	 HDR.DELTA_TIME := RIF_DELTA_TIME;
	 HDR.SEQ_NUM    := RIF_PACKETS_SENT;
	 B1.STATE := OBS_BUFFER_EMPTY;
	 DATA_INDEX := 0;
	 TM_EVENTS_PROCESSED := 0;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

	 if RIF_EVENTS_PACKED >= RIF_SORTED_EVENTS_NO then
            -- the whole image data has already been packed to the TM blocks
            -- terminate the sequence.
	    IASW_BUFFER.DISCARD_EVENTS(AMOUNT => RIF_EVENTS_NO);
	    HDR.SEQ_NUM := HDR.SEQ_NUM + 16#8000#;  -- signal the last packet
	    B1.STATE := OBS_SEQUENCE_COMPLETED;
	    exit;
	 elsif ((HDRLEN + (DATA_INDEX + 4 + 1) / 2) > K_MAX_TM_DATA_INDEX) then
	    -- 4 is the max. length in bytes of compressed event
            -- TM packet is ready to sent.
	    IASW_COMMON.INC(RIF_PACKETS_SENT);
	    B1.STATE := OBS_BUFFER_COMPLETED;
	    exit;
	 end if;

	 IASW_EVSORT.PEEK_SORTED_EVENT(
	    INDEX => RIF_EVENTS_PACKED,
	    EVENT => EVENT);
	 RIF_EVENTS_PACKED := RIF_EVENTS_PACKED + 1;

	 BB(0) := COORD_8BIT(EVENT(4),EVENT_X_TABLE);
	 BB(1) := COORD_8BIT(EVENT(3),EVENT_Y_TABLE);
	 if DATA_INDEX = 0 then
	    -- first event in the TM package
	    HDR.XY1_POS := W;    -- W is word equivalent to BB(0..1)
	    DELTA_POS := 0;
	 elsif W >= XY1 then
	    DELTA_POS := W - XY1;
	 else
	    -- fatal error - something wrong with sorting.
	    DELTA_POS := 16#FFFF#;
	 end if;
	 XY1 := W;
	 HDR.XY2_POS := XY1;

	 PEV.ENERGY := CCT.T_UNSIGNED_3(
            ENERGY_TABLE(IASW_COMMON.CVT(EVENT(2) mod 4096)) / 32);
	 if DELTA_POS < 31 then
	    PEV.DELTA_POSITION := CCT.T_UNSIGNED_5(DELTA_POS);
	    PUT_BYTE_TO_TM_DATA( BB(0), DATA_ADDR, DATA_INDEX );
	 else
	    PEV.DELTA_POSITION := 31;
	    PUT_BYTE_TO_TM_DATA( BB(0), DATA_ADDR, DATA_INDEX );
	    W := DELTA_POS;
	    if DELTA_POS < 255 then
	       PUT_BYTE_TO_TM_DATA( BB(1), DATA_ADDR, DATA_INDEX );
	    else
	       PUT_BYTE_TO_TM_DATA( 255,   DATA_ADDR, DATA_INDEX );
	       PUT_BYTE_TO_TM_DATA( BB(0), DATA_ADDR, DATA_INDEX );
	       PUT_BYTE_TO_TM_DATA( BB(1), DATA_ADDR, DATA_INDEX );
	    end if;
	 end if;

	 TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;
	 HDR.EVENTS          := TM_EVENTS_PROCESSED;
	 B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;
      end loop;
   end SCF_RESTRICTED;


   procedure SCF_SPECTRAL_TIMING is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT   : CIT.T_OBT;           -- OBT of the first event in the package.
	 NEV   : CCT.T_UNSIGNED_16;   -- number of events in the package
	 GF    : IASW_GF.T_GREY_FILTER_INDICATOR;  -- for the first event
      end record;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      type T_PACKED_WORD is record
         ENERGY     : T_UNSIGNED_8;
         DELTA_TIME : T_UNSIGNED_8;
      end record;

      for T_PACKED_WORD use record
         ENERGY     at 0 range 0 ..  7;
         DELTA_TIME at 0 range 8 .. 15;
      end record;

      function CVT is new UNCHECKED_CONVERSION
	(SOURCE => T_PACKED_WORD, TARGET => INTEGER);

      EVENT            : IASW_COMMON.T_EVENT; -- Event currently processed.
      DELTA_TIME       : CCT.T_UNSIGNED_16;   -- Time between events.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
      APPENDED         : BOOLEAN;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	 B1.INDEX := HDR'size / 16;
	 TM_EVENTS_PROCESSED := 0;
	 HDR.NEV             := TM_EVENTS_PROCESSED;
	 B1.STATE := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);
         FIXBUGG1_DO(EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

         if DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

            REQUEST_FOR_SCF_SWITCHING(EVENT(4));
            -- this will become effective only after completting the packet

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- no more events expected, exit the loop
            exit;

         else   -- normal event, or GF report

	    if B1.STATE = OBS_BUFFER_EMPTY then
	       -- Fill in the packet header for the first event.
               SET_SC_TM_DATA_HEADER( B1.DATA, 1,0,
                  IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), EVENT(1)));

	       HDR.GF := GREY_FILTER_REPORTED;

               if THIS_IS_START_OF_FORMAT then
                  -- without this the GET_DELTA_TIME(), which is near
                  -- below would have no PREV_xx for reference.
                  -- this is done only once at the beginning, so next time
                  -- the first event in the packet will have delta time
                  -- relative to the previous event (last one in the
                  -- previous packet), not zero.
                  PREV_CC := EVENT(0);
                  PREV_TC := EVENT(1);
                  THIS_IS_START_OF_FORMAT := FALSE;
               end if;
	    end if;

         end if;

	 APPENDED := FALSE;   -- preliminary

	 if DUMMY_EVENT_TYPE = IASW_COMMON.NORMAL_EVENT then

            -- delta time relative to the time of previous real event
	    DELTA_TIME := TRUNC16(GET_DELTA_TIME(
               EVENT(0), EVENT(1),
               PREV_CC, PREV_TC));

	    -- Append event to the packet.
	    if DELTA_TIME >= 255 then
	       B1.DATA(B1.INDEX + 1) := CVT(T_PACKED_WORD'(
                  ENERGY     => CCT.T_UNSIGNED_8(
                     ENERGY_TABLE(IASW_COMMON.CVT(EVENT(2) mod 4096))),
                  DELTA_TIME => 255));
	       B1.DATA(B1.INDEX + 2) := IASW_COMMON.CVT(DELTA_TIME);
	       B1.INDEX := B1.INDEX + 2;
	    else
	       B1.DATA(B1.INDEX + 1) := CVT(T_PACKED_WORD'(
                  ENERGY     => CCT.T_UNSIGNED_8(
                     ENERGY_TABLE(IASW_COMMON.CVT(EVENT(2) mod 4096))),
                  DELTA_TIME => CCT.T_UNSIGNED_8(DELTA_TIME)));
	       B1.INDEX := B1.INDEX + 1;
	    end if;

	    -- Store HKCC and TC for further use.
            -- only real events time is stored.
	    PREV_CC := EVENT(0);
	    PREV_TC := EVENT(1);

	    APPENDED := TRUE;

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

            ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));

	    B1.DATA(B1.INDEX + 1) := CVT(T_PACKED_WORD'(
               ENERGY     => 0,
               DELTA_TIME => CCT.T_UNSIGNED_8(GREY_FILTER_REPORTED)));
	    B1.INDEX := B1.INDEX + 1;

	    APPENDED := TRUE;

	 else
	    null;  -- another dummy event-record, skip it.
	 end if;

	 if APPENDED then
	    B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;

	    -- Increment the counter of events in the packet.
	    TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;
	    HDR.NEV             := TM_EVENTS_PROCESSED;

	    -- Check if this was last event in the packet.
	    if ((B1.INDEX + 2) > K_MAX_TM_DATA_INDEX) then
	       B1.STATE := OBS_SEQUENCE_COMPLETED;
	    end if;
	 end if;
      end loop;
   end SCF_SPECTRAL_TIMING;


   ------------------------------------------------------------------------
   -- SCF_TIMING
   --

   procedure PUT_W6_TO_TM_DATA(   -- local in SCF_TIMING
      D : in CCT.T_UNSIGNED_6;
      A : in SYSTEM.ADDRESS;
      I : in INTEGER;
      N : out INTEGER
      -- on return N shows words occupancy after (I+1)-th PUT.
   ) is
      type T_TIMING_8 is record
         D1,D2 : CCT.T_UNSIGNED_6;
         D3a   : CCT.T_UNSIGNED_4;
         D3b   : CCT.T_UNSIGNED_2;
         D4,D5 : CCT.T_UNSIGNED_6;
         D6a   : CCT.T_UNSIGNED_2;
         D6b   : CCT.T_UNSIGNED_4;
         D7,D8 : CCT.T_UNSIGNED_6;
      end record;
      for T_TIMING_8 use record
         D1  at 0 range 0..5;
         D2  at 0 range 6..11;
         D3a at 0 range 12..15;
         D3b at 1 range 0..1;
         D4  at 1 range 2..7;
         D5  at 1 range 8..13;
         D6a at 1 range 14..15;
         D6b at 2 range 0..3;
         D7  at 2 range 4..9;
         D8  at 2 range 10..15;
      end record;
      B : T_TIMING_8;
         for B use at A;
      -- this 'use at' can be not safe if the parameter A points to
      -- a wrong area.  The 3-words long area filled with this procedure
      -- coveres addresses A .. A+2.
   begin
      case I is
   when 1 =>
         B.D1 := D;
         N := 1;
   when 2 =>
         B.D2 := D;
         N := 2;
   when 3 =>
         B.D3a := CCT.T_UNSIGNED_4(D / 4);
         B.D3b := CCT.T_UNSIGNED_2(D mod 4);
         N := 2;
   when 4 =>
         B.D4 := D;
         N := 2;
   when 5 =>
         B.D5 := D;
         N := 3;
   when 6 =>
         B.D6a := CCT.T_UNSIGNED_2(D / 16);
         B.D6b := CCT.T_UNSIGNED_4(D mod 16);
         N := 3;
   when 7 =>
         B.D7 := D;
         N := 3;
   when 8 =>
         B.D8 := D;
         N := 1;
   when others =>
         N := 0;
         null;
      end case;
   end PUT_W6_TO_TM_DATA;


   procedure SCF_TIMING is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT   : CIT.T_OBT;           -- OBT of the first event in the package.
	 NEV   : CCT.T_UNSIGNED_16;   -- number of events in the package
	 GF    : IASW_GF.T_GREY_FILTER_INDICATOR;  -- for the first event
      end record;

      HDRLEN : constant INTEGER := T_HEADER'Size / 16;
      DATA_ADDR : SYSTEM.ADDRESS := B1.DATA(HDRLEN + 1)'address;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      BB : IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..1);
      DELTA_TIME : CCT.T_UNSIGNED_16;
         for DELTA_TIME use at BB'Address;
         -- this 'use at' is safe.

      NEXT_PUT_OCCUPANCY : INTEGER;
      EVENT      : IASW_COMMON.T_EVENT; -- Event currently processed.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
      APPENDED : BOOLEAN;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	 B1.INDEX  := HDRLEN;
         -- B1.INDEX+1 points to the beginning of a group of 8 data
         TIM_INDEX := 1;  -- inside a group of 8 data  (range 1..8)
         TIM_EXT_INDEX := K_MAX_TM_DATA_INDEX;
	 TM_EVENTS_PROCESSED := 0;
	 HDR.NEV             := TM_EVENTS_PROCESSED;
	 B1.STATE := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);
         FIXBUGG1_DO(EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

         if DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

            REQUEST_FOR_SCF_SWITCHING(EVENT(4));
            -- this will become effective only after completting the packet

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- no more events expected, exit the loop
            exit;

         else   -- normal event, or GF report

	    if B1.STATE = OBS_BUFFER_EMPTY then
	       -- Fill in the packet header for the first event.
               SET_SC_TM_DATA_HEADER( B1.DATA, 2,0,
                  IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), EVENT(1)));

	       HDR.GF := GREY_FILTER_REPORTED;

               if THIS_IS_START_OF_FORMAT then
                  -- without this the GET_DELTA_TIME(), which is near
                  -- below would have no PREV_xx for reference.
                  -- this is done only once at the beginning, so next time
                  -- the first event in the packet will have delta time
                  -- relative to the previous event (last one in the
                  -- previous packet), not zero.
                  PREV_CC := EVENT(0);
                  PREV_TC := EVENT(1);
                  THIS_IS_START_OF_FORMAT := FALSE;
               end if;
	    end if;

         end if;

	 APPENDED := FALSE;   --preliminary

	 if DUMMY_EVENT_TYPE = IASW_COMMON.NORMAL_EVENT then

            -- delta time relative to the time of previous real event
	    DELTA_TIME := TRUNC16(GET_DELTA_TIME(
               EVENT(0), EVENT(1),
               PREV_CC, PREV_TC));
            if DELTA_TIME > 16#FFDF# then
               DELTA_TIME := 16#FFDF#;
            end if;

	    -- Store HKCC and TC for further use.
            -- only real events time is stored.
	    PREV_CC := EVENT(0);
	    PREV_TC := EVENT(1);

	    APPENDED := TRUE;

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

            ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));

            DELTA_TIME := 16#FFE0# + CCT.T_UNSIGNED_16(GREY_FILTER_REPORTED);

	    APPENDED := TRUE;

	 else
	    null;  -- another dummy event-record, skip it.
	 end if;

	 if APPENDED then
	    -- Append event to the end of the packet.
	    if DELTA_TIME < 63 then
	       PUT_W6_TO_TM_DATA(
                 D => CCT.T_UNSIGNED_6(DELTA_TIME),
                 A => B1.DATA(B1.INDEX+1)'address,
                 I => TIM_INDEX,
                 N => NEXT_PUT_OCCUPANCY );
	    else
	       PUT_W6_TO_TM_DATA(
                 D => 63,
                 A => B1.DATA(B1.INDEX+1)'address,
                 I => TIM_INDEX,
                 N => NEXT_PUT_OCCUPANCY );
               B1.DATA(TIM_EXT_INDEX) := IASW_COMMON.CVT(DELTA_TIME);
               TIM_EXT_INDEX := TIM_EXT_INDEX - 1;
	    end if;
            if TIM_INDEX < 8 then
               TIM_INDEX := TIM_INDEX + 1;
            else
               B1.INDEX := B1.INDEX + 3;
               TIM_INDEX := 1;
            end if;

	    B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;

	    -- Increment the counter of events in the packet.
	    TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;
	    HDR.NEV             := TM_EVENTS_PROCESSED;

	    -- Check if this was last event in the packet.
            if (B1.INDEX + NEXT_PUT_OCCUPANCY + 1) > TIM_EXT_INDEX then
	       B1.STATE := OBS_SEQUENCE_COMPLETED;
	    end if;
	 end if;
      end loop;
   end SCF_TIMING;



   ------------------------------------------------------------------------
   -- SCF_SPECTRUM
   --

   procedure CLEAR_SPECTRUM is   -- local in SCF_SPECTRUM
   begin
      for I in SF_COUNTERS'Range loop
	 SF_COUNTERS(I) := 0;
      end loop;
   end CLEAR_SPECTRUM;


   procedure PUT4(   -- local in PACK_SPECTRUM
      A : in SYSTEM.ADDRESS;
      X : in CCT.T_UNSIGNED_4;
      I : INTEGER
   ) is
      type T_4X4 is array(0..3) of CCT.T_UNSIGNED_4;
         pragma pack(T_4X4);
      W1 : CCT.T_UNSIGNED_16;
      W4 : T_4X4;
         for W4 use at W1'address;
         -- this 'use at' is safe.
      WA : array(0..19) of CCT.T_UNSIGNED_16;
         for WA use at A;
         -- this 'use at' can be not safe if the parameters A points to
         -- a wrong place.  The procedure fills 20-words area pointed by A.
      IA,I4 : INTEGER;
   begin
      IA := I / 4;
      I4 := I mod 4;
      W1 := WA(IA);
      W4(I4) := X;
      WA(IA) := W1;
   end PUT4;

   procedure PUT8(   -- local in PACK_SPECTRUM
      A : in SYSTEM.ADDRESS;
      X : in CCT.T_UNSIGNED_8;
      I : INTEGER
   ) is
      type T_2X8 is array(0..1) of CCT.T_UNSIGNED_8;
         pragma pack(T_2X8);
      W1 : CCT.T_UNSIGNED_16;
      W2 : T_2X8;
         for W2 use at W1'address;
         -- this 'use at' is safe.
      WA : array(0..19) of CCT.T_UNSIGNED_16;
         for WA use at A;
         -- this 'use at' can be not safe if the parameters A points to
         -- a wrong place.  The procedure fills 20-words area pointed by A.
      IA,I2 : INTEGER;
   begin
      IA := I / 2;
      I2 := I mod 2;
      W1 := WA(IA);
      W2(I2) := X;
      WA(IA) := W1;
   end PUT8;


   procedure PACK_SPECTRUM(   -- local in SCF_SPECTRUM
      TM_DATA_ADDR : in SYSTEM.ADDRESS
   ) is
      type T_DATA_SET is record
         SPECTRUM : IASW_COMMON.T_ARRAY_OF_WORDS(0..15); --T_SPECTRUM;
         ADDS     : IASW_COMMON.T_ARRAY_OF_WORDS(0..3); --T_ADDS;
      end record;
      DATA : T_DATA_SET;
         for DATA use at TM_DATA_ADDR;
         -- this 'use at' can be not safe if the parameters TM_DATA_ADDR points to
         -- a wrong place.  The procedure fills 20-words area pointed by A.
      ADDS : INTEGER;
      I2   : INTEGER;
      CNT  : CCT.T_UNSIGNED_16;
   begin
      ADDS := 0;
      for I in 0..63 loop
	 I2 := I;       -- without this trick there was error
                        -- "OUT OF REGISTERS (SELREG)"
	 CNT := SF_COUNTERS(I2);
	 if CNT < 15 then
	    PUT4(DATA.SPECTRUM'address, CCT.T_UNSIGNED_4(CNT), I2);
	 else
	    PUT4(DATA.SPECTRUM'address, 15, I2);
	    if CNT > 255 then
	       PUT8(DATA.ADDS'address, 255, ADDS);
	    else
	       PUT8(DATA.ADDS'address, CCT.T_UNSIGNED_8(CNT), ADDS);
	    end if;
	    if ADDS < 7 then
	       ADDS := ADDS + 1;
	    end if;
	 end if;
      end loop;
   end PACK_SPECTRUM;


   procedure SCF_SPECTRUM is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT  : CIT.T_OBT;          -- OBT of the first event in the package.
	 NDAT : CCT.T_UNSIGNED_16;  -- number of data sets in the package
	 GF   : IASW_GF.T_GREY_FILTER_INDICATOR; -- for the first event
      end record;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      IDX : INTEGER;
      SPECTRUM_PERIOD : constant CCT.T_UNSIGNED_16 := 1024;  -- 125 ms
      EVENT : IASW_COMMON.T_EVENT;        -- Event currently processed.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
   begin
      while B1.STATE <= OBS_BUFFER_PARTIALLY_FILLED loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.PEEK_EVENT (INDEX => 0, EVENT => EVENT);
	 -- PEEK_EVENT is used instead of GET_EVENT because it gives
	 -- opportunity to leave the event in the buffer if it should be
	 -- used in the next TM package.
         -- used events must be removed from the buffer with DISCARD_EVENTS(1);
         FIXBUGG1_DO(EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

	 if DUMMY_EVENT_TYPE = IASW_COMMON.NORMAL_EVENT then

	    if B1.STATE <= OBS_BUFFER_NEED_INIT then
	       IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	       B1.INDEX := HDR'size / 16;
	       TM_EVENTS_PROCESSED := 0;
	       HDR.NDAT            := 0;
               SET_SC_TM_DATA_HEADER( B1.DATA, 2,8,
                  IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), CLEAR10BITS(EVENT(1))));
	       HDR.GF := GREY_FILTER_REPORTED;
	       SF_CC := EVENT(0);
	       SF_TC := EVENT(1) / SPECTRUM_PERIOD;
	       CLEAR_SPECTRUM;
	       B1.STATE := OBS_BUFFER_EMPTY;
	    end if;

	    -- Check if this event should be placed in the next spectrum.
	    -- If so, append current spectrum to TM buffer and
	    -- start counting over.
	    while ((EVENT(0) /= SF_CC) or
                  ((EVENT(1) / SPECTRUM_PERIOD) /= SF_TC)) loop

	       B1.INDEX := B1.INDEX + 1;
               B1.DATA(B1.INDEX) := IASW_COMMON.CVT(GREY_FILTER_REPORTED);

	       PACK_SPECTRUM (B1.DATA(B1.INDEX+1)'address);
               -- Warning! When changing something here be sure that
               -- the index (B1.INDEX+1) starts the area of at least 20 words
               -- in the array B1.DATA.

	       B1.INDEX := B1.INDEX + 20;
	       IASW_COMMON.INC (HDR.NDAT);
	       CLEAR_SPECTRUM;

	       -- Check if this was last spectrum in the packet.
	       if ((B1.INDEX + 21) > K_MAX_TM_DATA_INDEX) then
		  B1.STATE := OBS_SEQUENCE_COMPLETED;
                  -- this is the place, where DISCARD_EVENTS(1) is not used,
                  -- so this event will start the next TM packet.
                  FIXBUGG1_BACK;
		  return;
	       else
		  B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;
	       end if;

	       -- Move forward HK_CC and TC for the next spectrum
	       if SF_TC < 63 then  -- 63 = (64K/SPECTRUM_PERIOD)-1
		  IASW_COMMON.INC (SF_TC);
	       else
		  IASW_COMMON.INC (SF_CC, 1, 0);
		  SF_TC := 0;
	       end if;
	    end loop;

	    IDX := INTEGER(ENERGY_TABLE(INTEGER(EVENT(2) mod 4096)) / 4);
	    IASW_COMMON.INC (SF_COUNTERS(IDX));
	    IASW_BUFFER.DISCARD_EVENTS(1); -- see comment above to PEEK_EVENT()

	    -- count events processed.
	    TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.GREY_FILTER_REPORT then

            ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(EVENT(4));
	    IASW_BUFFER.DISCARD_EVENTS(1);

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.CHANGE_FORMAT_REQUEST then

            REQUEST_FOR_SCF_SWITCHING(EVENT(4));
	    IASW_BUFFER.DISCARD_EVENTS(1);
            -- this will become effective only after completting the packet

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
	    IASW_BUFFER.DISCARD_EVENTS(1);
            -- no more events expected, exit the loop
            exit;

	 else
	    -- Another dummy event-record, skip it.
	    IASW_BUFFER.DISCARD_EVENTS (1);
	 end if;
      end loop;
   end SCF_SPECTRUM;


   procedure SCF_CALIBRATION is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT      : CIT.T_OBT;      -- OBT of the first event in the package.
	 CAL_AMPL : CCT.T_UNSIGNED_16;  -- Calibrator amplitude.
	 CAL_FREQ : CCT.T_UNSIGNED_16;  -- Calibrator frequency.
	 ASS      : CCT.T_UNSIGNED_16;  -- Anode switch status.
      end record;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      JI : INTEGER;
      JW : CCT.T_UNSIGNED_16;
      HWS : IASW_DFEE.T_HW_SETTINGS;
      EVENT : IASW_COMMON.T_EVENT;      -- Event currently processed.
      DELTA_TIME : CCT.T_UNSIGNED_16;   -- Time between events.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	 B1.INDEX := HDR'size / 16;
	 TM_EVENTS_PROCESSED := 0;
	 HDR.CAL_AMPL := 0;
	 HDR.CAL_FREQ := 0;
	 IASW_DFEE.GET_HW_SETTINGS( HWS );
	 HDR.ASS := HWS.ANODE_SETTINGS;
	 B1.STATE := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);
         FIXBUGG1_DO(EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

	 if DUMMY_EVENT_TYPE = IASW_COMMON.CALIBRATION_AMPLITUDE_REPORT then

	    CF_AMPL := EVENT(4);
	    CF_FREQ := EVENT(5);

	 elsif DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then

            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- exceptionally here 'exit' is not required, because the
            -- End Of Data Marker should also be inserted into TM packet.
            -- exit;

	 end if;

	 if B1.STATE = OBS_BUFFER_EMPTY then
	    -- Fill in the packet header for the first data event.
            SET_SC_TM_DATA_HEADER( B1.DATA, 0,1,
               IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), EVENT(1)));
	    HDR.CAL_AMPL := CF_AMPL;
	    HDR.CAL_FREQ := CF_FREQ;
            if THIS_IS_START_OF_FORMAT then
               -- without this the GET_DELTA_TIME(), which is near
               -- below would have no PREV_xx for reference.
               -- this is done only once at the beginning, so next time
               -- the first event in the packet will have delta time
               -- relative to the previous event (last one in the
               -- previous packet), not zero.
               PREV_CC := EVENT(0);
               PREV_TC := EVENT(1);
               THIS_IS_START_OF_FORMAT := FALSE;
            end if;
	 end if;

         -- delta time relative to the time of previous event
         -- of any kind, including amplitude marker.
	 DELTA_TIME := TRUNC16(GET_DELTA_TIME(
            EVENT(0), EVENT(1),
            PREV_CC, PREV_TC));

	 -- Append event to the end of the packet.
	 B1.DATA(B1.INDEX + 1) := IASW_COMMON.CVT(DELTA_TIME);

	 for I in IASW_COMMON.T_EVENT_SIZE range 2 .. EVENT_SIZE-1 loop
	    JW := I;
	    JI := B1.INDEX + IASW_COMMON.CVT(JW);
	    B1.DATA (JI) := IASW_COMMON.CVT(EVENT (I));
	 end loop;

	 B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;

	 -- Increment the counter of events in the packet.
	 B1.INDEX := B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE)-1;
	 TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;

	 -- Store HKCC and TC for further use.
         -- differente than in other formats it is done for
         -- any kind of event.

	 --by eliminating next two lines delta-time will be calc. relative
	 --to the first event in the packet, not to the previous one.
         --if the two lines are eliminated then above setting for
         --PREV_CC and PREV_TC should be done independently from value
         --of the flag THIS_IS_START_OF_FORMAT.
	 PREV_CC := EVENT(0);
	 PREV_TC := EVENT(1);

	 if ((B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE)-1) > K_MAX_TM_DATA_INDEX) then
	    B1.STATE := OBS_SEQUENCE_COMPLETED;
	 end if;
      end loop;
   end SCF_CALIBRATION;


   procedure SCF_DIAGNOSTIC is
   use IASW_COMMON;   -- because of comparision of T_DUMMY_EVENT-s
      type T_HEADER is record
	 OBT      : CIT.T_OBT;     -- OBT of the first event in the package.
	 NEV      : CCT.T_UNSIGNED_16;   -- number of events in the package
	 SPARE2   : CCT.T_UNSIGNED_16;
	 ASS      : CCT.T_UNSIGNED_16;  -- Anode switch status.
      end record;

      HDR : T_HEADER;
         for HDR use at B1.DATA'address;
pragma If(((HDR'size+15)/16) > B1.DATA'length);
   error: Violated safety of 'use at'
pragma End if;

      JI : INTEGER;
      JW : CCT.T_UNSIGNED_16;
      HWS : IASW_DFEE.T_HW_SETTINGS;
      EVENT : IASW_COMMON.T_EVENT;      -- Event currently processed.
      DUMMY_EVENT_TYPE : IASW_COMMON.T_DUMMY_EVENT;
   begin
      if B1.STATE <= OBS_BUFFER_NEED_INIT then
	 IASW_TM.INIT_SCIENCE_TM (B1.DATA);
	 B1.INDEX := HDR'size / 16;
	 TM_EVENTS_PROCESSED := 0;
	 HDR.NEV := 0;
	 HDR.SPARE2 := 0;
	 IASW_DFEE.GET_HW_SETTINGS( HWS );
	 HDR.ASS := HWS.ANODE_SETTINGS;
	 B1.STATE := OBS_BUFFER_EMPTY;
      end if;

      while (B1.STATE in OBS_BUFFER_EMPTY .. OBS_BUFFER_PARTIALLY_FILLED) loop

         exit when NO_DATA_IN_THE_BUFFER;
	 IASW_BUFFER.GET_EVENT (EVENT);
	 DUMMY_EVENT_TYPE := IASW_COMMON.GET_DUMMY_EVENT_TYPE (EVENT);

	 if DUMMY_EVENT_TYPE = IASW_COMMON.END_OF_DATA_MARKER then
            SCTM_STOP(K_SCTM_SOFT_STOP);
            -- similarly like in calibration there is no need for 'exit' here.
            -- exit;
         end if;

	 if B1.STATE = OBS_BUFFER_EMPTY then
	    -- Fill in the packet header for the first data event.
            SET_SC_TM_DATA_HEADER( B1.DATA, 0,2,
               IASW_TIME.GET_OBT_OF_EVENT(EVENT(0), EVENT(1)));
	 end if;

	 -- Append event to the end of the packet.

	 for I in IASW_COMMON.T_EVENT_SIZE range 0 .. EVENT_SIZE-1 loop
	    JW := I;
	    JI := B1.INDEX + IASW_COMMON.CVT(JW) + 1;
	    B1.DATA(JI) := IASW_COMMON.CVT(EVENT (I));
	 end loop;

	 B1.STATE := OBS_BUFFER_PARTIALLY_FILLED;

	 -- Increment the counter of events in the packet.
	 B1.INDEX := B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE);
	 TM_EVENTS_PROCESSED := TM_EVENTS_PROCESSED + 1;
         HDR.NEV             := TM_EVENTS_PROCESSED;

	 if ((B1.INDEX + IASW_COMMON.CVT(EVENT_SIZE)) > K_MAX_TM_DATA_INDEX) then
	    B1.STATE := OBS_SEQUENCE_COMPLETED;
	 end if;
      end loop;
   end SCF_DIAGNOSTIC;


   procedure GET_AND_COMPRESS_EVENTS is   -- local in TASK_SC_OUT
   begin
      case SCF_ACTUAL is
   when IASW_MODES.K_SCF_TEST =>        SCF_TEST;
   when IASW_MODES.K_SCF_FULL =>        SCF_FULL;
   when IASW_MODES.K_SCF_RESTRICTED =>  SCF_RESTRICTED;
   when IASW_MODES.K_SCF_SPECT_TIM =>   SCF_SPECTRAL_TIMING;
   when IASW_MODES.K_SCF_TIMING =>      SCF_TIMING;
   when IASW_MODES.K_SCF_SPECTRUM  =>   SCF_SPECTRUM;
   when IASW_MODES.K_SCF_CALIBRATION => SCF_CALIBRATION;
   when IASW_MODES.K_SCF_DIAGNOSTIC =>  SCF_DIAGNOSTIC;
   when others =>                       SCF_UNDEFINED;
      end case;
   end GET_AND_COMPRESS_EVENTS;


   procedure SET_APID_FOR_THE_FORMAT(   -- local in SWITCH_TO_FORMAT
      SCF : in IASW_MODES.T_SCIENCE_FORMAT
   ) is
   begin
      case SCF is
   when IASW_MODES.K_SCF_TEST        => B1.APID := IASW_COMMON.APID_TM_SC_TEST;
   when IASW_MODES.K_SCF_FULL        => B1.APID := IASW_COMMON.APID_TM_SC_FULL;
   when IASW_MODES.K_SCF_RESTRICTED  => B1.APID := IASW_COMMON.APID_TM_SC_RESTRICTED;
   when IASW_MODES.K_SCF_SPECT_TIM   => B1.APID := IASW_COMMON.APID_TM_SC_SPECT_TIM;
   when IASW_MODES.K_SCF_TIMING      => B1.APID := IASW_COMMON.APID_TM_SC_TIMING;
   when IASW_MODES.K_SCF_SPECTRUM    => B1.APID := IASW_COMMON.APID_TM_SC_SPECTRUM;
   when IASW_MODES.K_SCF_CALIBRATION => B1.APID := IASW_COMMON.APID_TM_SC_CALIBRATION;
   when IASW_MODES.K_SCF_DIAGNOSTIC  => B1.APID := IASW_COMMON.APID_TM_SC_DIAGNOSTIC;
   when others => null;  -- e.g. IASW_MODES.K_SCF_UNDEFINED
      end case;
   end SET_APID_FOR_THE_FORMAT;


   procedure SWITCH_TO_FORMAT(   -- local in TASK_SC_OUT
      SCF : IASW_MODES.T_SCIENCE_FORMAT;
      EVS : IASW_COMMON.T_EVENT_SIZE
   ) is
   begin
      THIS_IS_START_OF_FORMAT := (SCF /= SCF_ACTUAL);
      CR_OUT;
      CR_INIT;
      B1.STATE := OBS_SEQUENCE_NEED_INIT;
      SET_APID_FOR_THE_FORMAT(SCF);
      SCF_ACTUAL := SCF;
      EVENT_SIZE := EVS;
   end SWITCH_TO_FORMAT;


   procedure TASK_SC_OUT is
   --
   -- Reads events from cyclic buffer, forms packets and sends them via TM.
   --
      DO_DELAY : BOOLEAN;
   begin
      loop
	 begin
	    IASW_COMMON.I_AM_ALIVE(IASW_COMMON.TASK_SC_OUT_NO);

            -- is there a stop request?
            case SCTM_STOP_TYPE is
         when K_SCTM_HARD_STOP =>
               SCTM_STOP_TYPE := K_SCTM_NO_STOP;
	       COMPRESSION_PROC_PHASE := CPP_NOOP;

         when K_SCTM_SOFT_STOP =>
               SCTM_STOP_TYPE := K_SCTM_NO_STOP;
               if COMPRESSION_PROC_PHASE > CPP_SOFT_STOP then
                  COMPRESSION_PROC_PHASE := CPP_SOFT_STOP;
               end if;

         when K_SCTM_DELAYED_SOFT_STOP =>
               if COMPRESSION_PROC_PHASE > CPP_SOFT_STOP then
                  if SOFT_STOP_DELAY >= 2 then
                     SCTM_STOP_TYPE := K_SCTM_NO_STOP;
                     COMPRESSION_PROC_PHASE := CPP_SOFT_STOP;
                  end if;
               end if;

         when others =>
               null;
            end case;

	    -- Get events from the buffer
	    if COMPRESSION_PROC_PHASE > CPP_NOOP then
	       GET_AND_COMPRESS_EVENTS;
	    end if;

	    -- Terminate if there is no more data, and no hope
            -- for data to come, in soft stop mode
	    if (COMPRESSION_PROC_PHASE = CPP_SOFT_STOP) and
	       IASW_BUFFER.IS_BUFFER_EMPTY and
	       (B1.STATE < OBS_BUFFER_COMPLETED)
            then
               CR_OUT;
               COMPRESSION_PROC_PHASE := CPP_NOOP;
	    end if;

	    -- Produce TM output
            DO_DELAY := TRUE;   -- preliminary
	    if B1.STATE >= OBS_BUFFER_COMPLETED then
	       if DPE_CSSW.GET_ICB_STATE(CIT.TM_SCI).FREE >= TM_MIN_FREE_ICB then
	          SC_TM_OUT (B1);

                  if (B1.STATE <= OBS_BUFFER_NEED_INIT) then
                     -- this means that the packet has been successfuly sent.

		     -- Count output throughput in [events/BCP1]
		     IASW_COMMON.INC(
		        X         => TM_RATE_CNT,
		        INCREMENT => TM_EVENTS_PROCESSED);

                     -- shall we delay?
                     GOOD_TM_CNT := (GOOD_TM_CNT + 1) mod 3;
                     DO_DELAY := (GOOD_TM_CNT = 0);
                  end if;
               end if;
            end if;

            -- switch to another format, if required and possible
            if SCF_SWITCH_REQUEST and
               (COMPRESSION_PROC_PHASE > CPP_NOOP) and
               (B1.STATE <= OBS_SEQUENCE_NEED_INIT)
            then
               SWITCH_TO_FORMAT(SCF_WANTED, EVSIZE_WANTED);
               SCF_SWITCH_REQUEST := FALSE;
            end if;

            -- give other tasks chance to work
            if DO_DELAY then
               DPE_CSSW.DELAY_RELATIVE (1);
            end if;

	 exception
	    when IASW_BUFFER.EX_INDEX_OUT_OF_RANGE |
		 IASW_BUFFER.EX_BUFFER_EMPTY =>
	       IASW_TM.PUT_OEM (IASW_TM.EV_IASW_BUFFER, DPE_CSSW.CURRENT_TASK);

	    when others =>
	       IASW_TM.PUT_OEM (IASW_TM.EV_IASW_EXCEPTION,
				DPE_CSSW.CURRENT_TASK);
	 end;
      end loop;
   end TASK_SC_OUT;


   function IS_ENERGY_TABLE_VALID(
   --
   -- Checks if data describing energy table is valid.  The data is valid
   -- when the energy values are stored in the increasing order, and the last one
   -- is equal to 4095.
   --
      DATA      : in IASW_COMMON.T_COMPRESSED_ENERGY_TABLE
   ) return BOOLEAN is
   begin
      if DATA(DATA'last) /= 4095 then
	 return FALSE;
      end if;

      for I in DATA'first+1 .. DATA'last loop
	 if DATA(I) <= DATA(I-1) then
	    return FALSE;
	 end if;
      end loop;

      return TRUE;
   end IS_ENERGY_TABLE_VALID;


   procedure DECOMPRESS_ENERGY_TABLE(
   --
   -- Decompresses energy table from the compressed form (padded to
   -- SET_SW_PARAM routine).  The data should be checked first against its
   -- validity with IS_ENERGY_TABLE_VALID function.
   --
      DATA      : in  IASW_COMMON.T_COMPRESSED_ENERGY_TABLE;
      TABLE     : out T_ENERGY_TABLE
   ) is
      J      : NATURAL;                 -- Index to DATA table.
      ENERGY : CCT.T_UNSIGNED_16;        -- Current energy value.
   begin
      J      := DATA'first;
      ENERGY := 0;
      for I in 0..4095 loop
	 TABLE(I) := ENERGY;
	 if (I = IASW_COMMON.CVT(DATA(J))) and (J < DATA'last) then
	    J      := J      + 1;
	    ENERGY := ENERGY + 1;
	 end if;
      end loop;
   end DECOMPRESS_ENERGY_TABLE;


   procedure SET_ENERGY_TABLE (
      PARAM_VAL : in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN : in IASW_COMMON.T_SW_PARAM_LEN
   ) is
      A : INTEGER;
   begin
      if PARAM_LEN < 2 then
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;
      A := IASW_COMMON.CVT(PARAM_VAL(1));
      if (A < 0) or (A >= 256) or ((A+PARAM_LEN-1) > 256) then
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_VALUE;
      end if;

      for i in 2..PARAM_LEN loop
	 COMPRESSED_ENERGY_TABLE(A+i-1) := PARAM_VAL(i);
      end loop;
   end SET_ENERGY_TABLE;


   procedure EXEC_ENERGY_TABLE is
   begin
      if not IS_ENERGY_TABLE_VALID(COMPRESSED_ENERGY_TABLE) then
         -- return to default
	 COMPRESSED_ENERGY_TABLE := DEFAULT_COMPRESSED_ENERGY_TABLE;
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_VALUE;
      end if;
      DECOMPRESS_ENERGY_TABLE(COMPRESSED_ENERGY_TABLE, ENERGY_TABLE);
      IASW_DFEE.SET_ENERGY_TABLE(COMPRESSED_ENERGY_TABLE);
      IASW_DFEE.EXEC_ENERGY_TABLE;
   end EXEC_ENERGY_TABLE;




   function IS_EVENT_POSITION_TABLE_VALID(
      DATA      : in T_COMPRESSED_EVENT_POSITION_TABLE
   ) return BOOLEAN is
   begin
      if DATA(DATA'first) /= 0 then
	 return FALSE;
      end if;

      if DATA(DATA'last) /= 1023 then
	 return FALSE;
      end if;

      for I in DATA'first+1 .. DATA'last loop
	 if DATA(I) <= DATA(I-1) then
	    return FALSE;
	 end if;
      end loop;

      return TRUE;
   end IS_EVENT_POSITION_TABLE_VALID;


   procedure DECOMPRESS_EVENT_POSITION_TABLE(
      DATA      : in  T_COMPRESSED_EVENT_POSITION_TABLE;
      TABLE     : out T_EVENT_POSITION_TABLE
   ) is
      J  : NATURAL;                 -- Index to DATA table.
      XY : CCT.T_UNSIGNED_8;        -- Current XY value.
   begin
      J  := DATA'first;
      XY := 0;
      for I in 0..1023 loop
         if XY = 0 then
            TABLE(I) := 1;
         else
            TABLE(I) := XY;
         end if;
	 if (I = IASW_COMMON.CVT(DATA(J))) and (J < DATA'last) then
	    J  := J  + 1;
	    XY := XY + 1;
	 end if;
      end loop;
   end DECOMPRESS_EVENT_POSITION_TABLE;


   procedure SET_EVENT_POSITION_TABLE (
      PARAM_VAL : in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN : in IASW_COMMON.T_SW_PARAM_LEN;
      COMPRESSED_EVENT_POSITION_TABLE : out T_COMPRESSED_EVENT_POSITION_TABLE
   ) is
      A : INTEGER;
   begin
      if PARAM_LEN < 2 then
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;
      A := IASW_COMMON.CVT(PARAM_VAL(1));
      if (A < 0) or (A >= 256) or ((A+PARAM_LEN-1) > 256) then
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_VALUE;
      end if;

      for i in 2..PARAM_LEN loop
	 COMPRESSED_EVENT_POSITION_TABLE(A+i-2) := PARAM_VAL(i);
      end loop;
   end SET_EVENT_POSITION_TABLE;


   procedure EXEC_EVENT_POSITION_TABLE(
      COMPRESSED_EVENT_POSITION_TABLE : in out T_COMPRESSED_EVENT_POSITION_TABLE;
      EVENT_POSITION_TABLE : out T_EVENT_POSITION_TABLE;
      DEFAULT_COMPRESSED_EVENT_POSITION_TABLE : in T_COMPRESSED_EVENT_POSITION_TABLE
   ) is
   begin
      if not IS_EVENT_POSITION_TABLE_VALID(COMPRESSED_EVENT_POSITION_TABLE) then
         -- return to default
	 COMPRESSED_EVENT_POSITION_TABLE := DEFAULT_COMPRESSED_EVENT_POSITION_TABLE;
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_VALUE;
      end if;
      DECOMPRESS_EVENT_POSITION_TABLE(COMPRESSED_EVENT_POSITION_TABLE, EVENT_POSITION_TABLE);
   end EXEC_EVENT_POSITION_TABLE;




   procedure SET_SW_PARAM(
   -- called on reception of TC(5,3) with TID = 4
      PARAM_ID : in IASW_COMMON.T_FID;
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN
   ) is
   begin
      if PARAM_VAL'length < PARAM_LEN then
	 raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;

      case PARAM_ID is
   when 43 =>
         IASW_GF.SET_GREY_FILTER (PARAM_VAL, PARAM_LEN);

   when 44 =>
         SET_ENERGY_TABLE (PARAM_VAL, PARAM_LEN);

   when 45 =>
         EXEC_ENERGY_TABLE;

   when 51 =>
         SET_EVENT_POSITION_TABLE( PARAM_VAL, PARAM_LEN,
            COMPRESSED_EVENT_X_TABLE);

   when 52 =>
         EXEC_EVENT_POSITION_TABLE(
            COMPRESSED_EVENT_X_TABLE,
            EVENT_X_TABLE,
            DEFAULT_COMPRESSED_EVENT_X_TABLE);

   when 53 =>
         SET_EVENT_POSITION_TABLE( PARAM_VAL, PARAM_LEN,
            COMPRESSED_EVENT_Y_TABLE);

   when 54 =>
         EXEC_EVENT_POSITION_TABLE(
            COMPRESSED_EVENT_Y_TABLE,
            EVENT_Y_TABLE,
            DEFAULT_COMPRESSED_EVENT_Y_TABLE);

   when 46 =>
         IASW_GF.SET_BUFFER_LEVELS(PARAM_VAL, PARAM_LEN);

   when 47 =>
         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         if PARAM_VAL(1) not in 1..IASW_EVSORT.MAX_AMOUNT_OF_EVENTS_TO_SORT then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_VALUE;
         end if;
         RIF_MAX_EVENTS := PARAM_VAL(1);

   when others =>
         raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end case;
   end SET_SW_PARAM;


   procedure GET_SW_PARAM(
   -- called on reception of TC(5,4) with TID = 4
      PARAM_ID : in  IASW_COMMON.T_FID;
      PARAM_VAL: out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN: out IASW_COMMON.T_SW_PARAM_LEN
   ) is
   begin
      case PARAM_ID is
	 when 46 =>
            IASW_GF.GET_BUFFER_LEVELS(PARAM_VAL, PARAM_LEN);

	 when others =>
	    raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end case;
   end GET_SW_PARAM;


   procedure START_DATA_TAKING is
   begin
      COMPRESSION_PROC_PHASE := CPP_NOOP;
      TM_RATE     := 0;
      TM_RATE_CNT := 0;
      CF_AMPL := 0;
      CF_FREQ := 0;
      GREY_FILTER_REPORTED := IASW_GF.GREY_FILTER_100;
      EVENT_SIZE := 0;
      SCF_WANTED := IASW_MODES.K_SCF_UNDEFINED;
      -- EVENT_SIZE and SCF_WANTED will come from task HSL via big buffer
      -- in form of dummy event.
      SWITCH_TO_FORMAT(SCF_WANTED, EVENT_SIZE);
      SCTM_STOP_TYPE := K_SCTM_NO_STOP;
      COMPRESSION_PROC_PHASE := CPP_NORMAL;
      RIF_TIME := 0;
      FIXBUGG1_INIT;
   end START_DATA_TAKING;


   function STOPPED return BOOLEAN is
   begin
      return (COMPRESSION_PROC_PHASE = CPP_NOOP);
   end STOPPED;


   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE) is
   begin
      null;
   end GET_DPE_STATE;


   procedure BCP1_TICK is
   begin
      TM_RATE := TM_RATE_CNT;
      TM_RATE_CNT := 0;
      IASW_COMMON.INC(SOFT_STOP_DELAY);
      case SCF_ACTUAL is
   when IASW_MODES.K_SCF_TEST =>        null;
   when IASW_MODES.K_SCF_FULL =>        null;
   when IASW_MODES.K_SCF_RESTRICTED =>  IASW_COMMON.INC(RIF_TIME);
   when IASW_MODES.K_SCF_SPECT_TIM =>   null;
   when IASW_MODES.K_SCF_TIMING =>      null;
   when IASW_MODES.K_SCF_SPECTRUM  =>   null;
   when IASW_MODES.K_SCF_CALIBRATION => null;
   when IASW_MODES.K_SCF_DIAGNOSTIC =>  null;
   when others =>                       null;
      end case;
   end;


   procedure CONSTRUCTOR is
   begin

      TM_EVENTS_PROCESSED := 0;
      TM_RATE_CNT         := 0;
      TM_RATE             := 0;
      GOOD_TM_CNT := 0;

      EVENT_SIZE := 0;
      SCF_WANTED := IASW_MODES.K_SCF_UNDEFINED;
      SCF_ACTUAL := IASW_MODES.K_SCF_UNDEFINED;
      SCF_SWITCH_REQUEST := FALSE;

      COMPRESSION_PROC_PHASE := CPP_NOOP;
      B1.STATE := OBS_SEQUENCE_NEED_INIT;
      B2.STATE := OBS_SEQUENCE_NEED_INIT;
      SCTM_STOP_TYPE := K_SCTM_NO_STOP;
      GREY_FILTER_REPORTED    := IASW_GF.GREY_FILTER_100;

      -- Set energy conversion tables to defaults.
      COMPRESSED_ENERGY_TABLE := DEFAULT_COMPRESSED_ENERGY_TABLE;
      DECOMPRESS_ENERGY_TABLE(
	 DATA      => COMPRESSED_ENERGY_TABLE,
	 TABLE     => ENERGY_TABLE);
      IASW_DFEE.SET_DEFAULT_ENERGY_TABLE(COMPRESSED_ENERGY_TABLE);

      COMPRESSED_EVENT_X_TABLE := DEFAULT_COMPRESSED_EVENT_X_TABLE;
      DECOMPRESS_EVENT_POSITION_TABLE(
	 DATA      => COMPRESSED_EVENT_X_TABLE,
	 TABLE     => EVENT_X_TABLE);

      COMPRESSED_EVENT_Y_TABLE := DEFAULT_COMPRESSED_EVENT_Y_TABLE;
      DECOMPRESS_EVENT_POSITION_TABLE(
	 DATA      => COMPRESSED_EVENT_Y_TABLE,
	 TABLE     => EVENT_Y_TABLE);

      CR_INIT;

   end CONSTRUCTOR;

end IASW_SC_OUT;
