
with SYSTEM;
with CSSW_COMMON_TYPES;  use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;
with IASW_COMMON;
with IASW_MODES;
with IASW_BUFFER;
with IASW_EVSORT;
with IASW_GF;

package IASW_SC_OUT is

   package CIT renames CSSW_IF_TYPES;
   package CCT renames CSSW_COMMON_TYPES;

   type T_SCTM_STOP_TYPE is (
      K_SCTM_NO_STOP,
      K_SCTM_DELAYED_SOFT_STOP,
      K_SCTM_SOFT_STOP,
      K_SCTM_HARD_STOP
   );

   --------------------------------------------------------------------
   -- Local types and data, made global because of possible patching --
   --------------------------------------------------------------------

   -- The task will leave free at least 2 entries to the output TM-queue
   TM_MIN_FREE_ICB : constant NATURAL := 2;

   -- TM throughput data.  Events rate unit is [events/BCP1 period]
   -- At successfull sending ot TM packet, TM_RATE_CNT is incremented.
   -- At BCP1, TM_RATE_CNT is copied to TM_RATE and next cleared.
   TM_RATE, TM_RATE_CNT : CCT.T_UNSIGNED_16;

   -- counter incremented modulo 3 every successfull TM sent.  it's value
   -- decides about using or not of DELAY_RELATIVE(1) after PUT_TM().
   -- thanks to this DPE will be able theoretically generate up to 192
   -- TM packets per BCP1 (EID-B requires 184).
   GOOD_TM_CNT : CCT.T_UNSIGNED_16;

   -- Two TM buffers: B1 - standard Science Format (primary
   -- or secondary),  B2 - for Count Rates format.

   type T_OUT_BUF_STATUS is (
      OBS_SEQUENCE_NEED_INIT,
      OBS_BUFFER_NEED_INIT,
      OBS_BUFFER_EMPTY,
      OBS_BUFFER_PARTIALLY_FILLED,
      OBS_BUFFER_COMPLETED,            -- This will be used only in R.Im.
      OBS_SEQUENCE_COMPLETED           -- This will normaly be used
   );

   type T_OUT_BUF is record
      STATE      : T_OUT_BUF_STATUS;
      APID       : CIT.T_IASW_APID;
      DATA       : CIT.T_SCIENCE_DATA_FIELD;
      INDEX      : NATURAL;     -- used for indexing DATA().
   end record;

   -- max allowed index of TM packet (-1, because of CRC field).
   K_MAX_TM_DATA_INDEX : constant INTEGER := CIT.T_SCIENCE_DATA_FIELD'Last-1;

   B1,B2 : T_OUT_BUF;    -- main buffer and CR buffer


   -- Counts events in one TM package
   TM_EVENTS_PROCESSED : CCT.T_UNSIGNED_16;

   -- data for calculating delta time between events
   PREV_CC : CCT.T_UNSIGNED_16;
   PREV_TC : CCT.T_UNSIGNED_16;
   THIS_IS_START_OF_FORMAT : BOOLEAN;

   -- data for fixing the bugg in DFEE - sometimes TC falls down to zero,
   -- but CC is late with incrementing.
   type T_FIXBUGG1_DATA is record
      CC, PREV_CC : CCT.T_UNSIGNED_16;
      VALID   : BOOLEAN;
   end record;
   FIXBUGG1 : T_FIXBUGG1_DATA;


   -- Data controling compression process.
   type T_COMPRESSION_PROCESS_PHASE is (
      CPP_NOOP,
      CPP_SOFT_STOP,
      CPP_NORMAL
   );
   COMPRESSION_PROC_PHASE : T_COMPRESSION_PROCESS_PHASE;
   EVENT_SIZE : IASW_COMMON.T_EVENT_SIZE;
   SCF_ACTUAL : IASW_MODES.T_SCIENCE_FORMAT;
   SCTM_STOP_TYPE : T_SCTM_STOP_TYPE;
   SOFT_STOP_DELAY : CCT.T_UNSIGNED_16;  -- timer counter, counts BCP1s

   -- data oryginating from dummy-events
   GREY_FILTER_REPORTED : IASW_GF.T_GREY_FILTER_INDICATOR;
   EVSIZE_WANTED : IASW_COMMON.T_EVENT_SIZE;
   SCF_WANTED : IASW_MODES.T_SCIENCE_FORMAT;
   SCF_SWITCH_REQUEST : BOOLEAN;


   -- global data used in Restricted Imaging Format

   RIF_MAX_EVENTS : IASW_BUFFER.T_BUFFER_INDEX := 
                    IASW_EVSORT.MAX_AMOUNT_OF_EVENTS_TO_SORT - 20;   -- =2580

   -- Number of events really considered ( <= RIF_EVENTS)
   RIF_EVENTS_NO : IASW_BUFFER.T_BUFFER_INDEX;

   -- Number of true events sorted ( <= RIF_EVENTS_NO )
   RIF_SORTED_EVENTS_NO : IASW_BUFFER.T_BUFFER_INDEX;

   -- Amount of events packed to TM blocks
   RIF_EVENTS_PACKED : IASW_BUFFER.T_BUFFER_INDEX;

   -- Amount of packets already sent.
   RIF_PACKETS_SENT  : CCT.T_UNSIGNED_16;

   RIF_OBT           : CIT.T_OBT;
   RIF_DELTA_TIME    : CIT.T_U32_BITS;
   RIF_GF1           : IASW_GF.T_GREY_FILTER_INDICATOR;
   RIF_GF2           : IASW_GF.T_GREY_FILTER_INDICATOR;

   -- time of collecting data for an image [# of BCP1's]
   RIF_TIME          : CCT.T_UNSIGNED_16;


   -- global Count-rate data

   -- 401 = 1 word + 400 bytes
   CR_TM_CHANNELS     : constant INTEGER := 401; -- Number of CR counters.

   CR_TM_GF_CHANNELS  : constant INTEGER := 8;   -- Number of GF slotts.

   -- type for counter of events in a time slott
   subtype T_CR_DATA is CCT.T_UNSIGNED_16;

   type T_GF_AND_CHANNEL is record
      GFI : CCT.T_UNSIGNED_5;    -- grey filter indicator (0..31)
      CRI : CCT.T_UNSIGNED_11;   -- time slott index (0..400)
   end record;
   for T_GF_AND_CHANNEL use record
      GFI at 0 range 0..4;
      CRI at 0 range 5..15;
   end record;

   CR_CC0, CR_TC0 : CCT.T_UNSIGNED_16;  -- CC & TC for first channel.
   CR_DATA     : array(0..CR_TM_CHANNELS    + 10 - 1) of T_CR_DATA;
   CR_GF_DATA  : array(1..CR_TM_GF_CHANNELS + 3     ) of T_GF_AND_CHANNEL;
   CR_INDEX    : INTEGER;               -- Index of last used counter.
   CR_GF_INDEX : INTEGER;               -- Index of last used GF slot.


   -- global data for SPECTRUM Im. F.

   SF_COUNTERS  : array (0..63) of CCT.T_UNSIGNED_16;
   SF_CC, SF_TC : CCT.T_UNSIGNED_16;


   -- global data used in SCF_CALIBRATION

   CF_AMPL : CCT.T_UNSIGNED_16;
   CF_FREQ : CCT.T_UNSIGNED_16;


   -- global data used in TIMING format

   TIM_INDEX, TIM_EXT_INDEX : INTEGER;


   -- Default buffer levels in % (between 0 and 100).

   DEFAULT_EXTRA_LOW_LEVEL   : constant CCT.T_UNSIGNED_16 := 5;
   DEFAULT_NORMAL_LOW_LEVEL  : constant CCT.T_UNSIGNED_16 := 20;
   DEFAULT_NORMAL_HIGH_LEVEL : constant CCT.T_UNSIGNED_16 := 50;
   DEFAULT_EXTRA_HIGH_LEVEL  : constant CCT.T_UNSIGNED_16 := 80;

   -- Current buffer levels in % (between 0 and 100).

   EXTRA_LOW_LEVEL,
   NORMAL_LOW_LEVEL,
   NORMAL_HIGH_LEVEL,
   EXTRA_HIGH_LEVEL
      : CCT.T_UNSIGNED_16;

   -- The buffer levels above converted to the positions in buffer.

   EXTRA_LOW_EVENT_THRESHOLD,
   NORMAL_LOW_EVENT_THRESHOLD,
   NORMAL_HIGH_EVENT_THRESHOLD,
   EXTRA_HIGH_EVENT_THRESHOLD
      : IASW_BUFFER.T_BUFFER_INDEX;



   -- Data types for event X/Y linearisation

   type T_EVENT_POSITION_TABLE is array(INTEGER range 0..1023) of CCT.T_UNSIGNED_8;
   subtype T_COMPRESSED_EVENT_POSITION_TABLE is IASW_COMMON.T_ARRAY_OF_WORDS(0..255);

pragma If(CIT.K_JEMX = CIT.K_JEMX_1);

   DEFAULT_COMPRESSED_EVENT_X_TABLE :
      constant T_COMPRESSED_EVENT_POSITION_TABLE := (
0, 5, 9, 13, 17, 21, 25, 29, 33, 37, 41, 45, 49, 53, 57, 61,
65, 69, 73, 77, 81, 85, 89, 93, 97, 101, 105, 109, 113, 117, 121, 125,
129, 133, 137, 141, 145, 149, 153, 157, 161, 165, 169, 173, 177, 181, 185, 189,
193, 197, 201, 205, 209, 213, 217, 221, 225, 229, 233, 237, 241, 245, 249, 253,
257, 261, 265, 269, 273, 277, 281, 285, 289, 293, 297, 301, 305, 309, 313, 317,
321, 325, 329, 333, 337, 341, 345, 349, 353, 357, 361, 365, 369, 373, 377, 381,
385, 389, 393, 397, 401, 405, 409, 413, 417, 421, 425, 429, 433, 437, 441, 445,
449, 453, 457, 461, 465, 469, 473, 477, 481, 485, 489, 493, 497, 501, 505, 509,
513, 517, 521, 525, 529, 533, 537, 541, 545, 549, 553, 557, 561, 565, 569, 573,
577, 581, 585, 589, 593, 597, 601, 605, 609, 613, 617, 621, 625, 629, 633, 637,
641, 645, 649, 653, 657, 661, 665, 669, 673, 677, 681, 685, 689, 693, 697, 701,
705, 709, 713, 717, 721, 725, 729, 733, 737, 741, 745, 749, 753, 757, 761, 765,
769, 773, 777, 781, 785, 789, 793, 797, 801, 805, 809, 813, 817, 821, 825, 829,
833, 837, 841, 845, 849, 853, 857, 861, 865, 869, 873, 877, 881, 885, 889, 893,
897, 901, 905, 909, 913, 917, 921, 925, 929, 933, 937, 941, 945, 949, 953, 957,
961, 965, 969, 973, 977, 981, 985, 989, 993, 997, 1001, 1005, 1009, 1013, 1017, 1023
   );
   DEFAULT_COMPRESSED_EVENT_Y_TABLE :
      constant T_COMPRESSED_EVENT_POSITION_TABLE := (
0, 5, 9, 13, 17, 21, 25, 29, 33, 37, 41, 45, 49, 53, 57, 61,
65, 69, 73, 77, 81, 85, 89, 93, 97, 101, 105, 109, 113, 117, 121, 125,
129, 133, 137, 141, 145, 149, 153, 157, 161, 165, 169, 173, 177, 181, 185, 189,
193, 197, 201, 205, 209, 213, 217, 221, 225, 229, 233, 237, 241, 245, 249, 253,
257, 261, 265, 269, 273, 277, 281, 285, 289, 293, 297, 301, 305, 309, 313, 317,
321, 325, 329, 333, 337, 341, 345, 349, 353, 357, 361, 365, 369, 373, 377, 381,
385, 389, 393, 397, 401, 405, 409, 413, 417, 421, 425, 429, 433, 437, 441, 445,
449, 453, 457, 461, 465, 469, 473, 477, 481, 485, 489, 493, 497, 501, 505, 509,
513, 517, 521, 525, 529, 533, 537, 541, 545, 549, 553, 557, 561, 565, 569, 573,
577, 581, 585, 589, 593, 597, 601, 605, 609, 613, 617, 621, 625, 629, 633, 637,
641, 645, 649, 653, 657, 661, 665, 669, 673, 677, 681, 685, 689, 693, 697, 701,
705, 709, 713, 717, 721, 725, 729, 733, 737, 741, 745, 749, 753, 757, 761, 765,
769, 773, 777, 781, 785, 789, 793, 797, 801, 805, 809, 813, 817, 821, 825, 829,
833, 837, 841, 845, 849, 853, 857, 861, 865, 869, 873, 877, 881, 885, 889, 893,
897, 901, 905, 909, 913, 917, 921, 925, 929, 933, 937, 941, 945, 949, 953, 957,
961, 965, 969, 973, 977, 981, 985, 989, 993, 997, 1001, 1005, 1009, 1013, 1017, 1023
   );

pragma End if;
pragma If(CIT.K_JEMX = CIT.K_JEMX_2);

   DEFAULT_COMPRESSED_EVENT_X_TABLE :
      constant T_COMPRESSED_EVENT_POSITION_TABLE := (
0, 5, 9, 13, 17, 21, 25, 29, 33, 37, 41, 45, 49, 53, 57, 61,
65, 69, 73, 77, 81, 85, 89, 93, 97, 101, 105, 109, 113, 117, 121, 125,
129, 133, 137, 141, 145, 149, 153, 157, 161, 165, 169, 173, 177, 181, 185, 189,
193, 197, 201, 205, 209, 213, 217, 221, 225, 229, 233, 237, 241, 245, 249, 253,
257, 261, 265, 269, 273, 277, 281, 285, 289, 293, 297, 301, 305, 309, 313, 317,
321, 325, 329, 333, 337, 341, 345, 349, 353, 357, 361, 365, 369, 373, 377, 381,
385, 389, 393, 397, 401, 405, 409, 413, 417, 421, 425, 429, 433, 437, 441, 445,
449, 453, 457, 461, 465, 469, 473, 477, 481, 485, 489, 493, 497, 501, 505, 509,
513, 517, 521, 525, 529, 533, 537, 541, 545, 549, 553, 557, 561, 565, 569, 573,
577, 581, 585, 589, 593, 597, 601, 605, 609, 613, 617, 621, 625, 629, 633, 637,
641, 645, 649, 653, 657, 661, 665, 669, 673, 677, 681, 685, 689, 693, 697, 701,
705, 709, 713, 717, 721, 725, 729, 733, 737, 741, 745, 749, 753, 757, 761, 765,
769, 773, 777, 781, 785, 789, 793, 797, 801, 805, 809, 813, 817, 821, 825, 829,
833, 837, 841, 845, 849, 853, 857, 861, 865, 869, 873, 877, 881, 885, 889, 893,
897, 901, 905, 909, 913, 917, 921, 925, 929, 933, 937, 941, 945, 949, 953, 957,
961, 965, 969, 973, 977, 981, 985, 989, 993, 997, 1001, 1005, 1009, 1013, 1017, 1023
   );
   DEFAULT_COMPRESSED_EVENT_Y_TABLE :
      constant T_COMPRESSED_EVENT_POSITION_TABLE := (
0, 5, 9, 13, 17, 21, 25, 29, 33, 37, 41, 45, 49, 53, 57, 61,
65, 69, 73, 77, 81, 85, 89, 93, 97, 101, 105, 109, 113, 117, 121, 125,
129, 133, 137, 141, 145, 149, 153, 157, 161, 165, 169, 173, 177, 181, 185, 189,
193, 197, 201, 205, 209, 213, 217, 221, 225, 229, 233, 237, 241, 245, 249, 253,
257, 261, 265, 269, 273, 277, 281, 285, 289, 293, 297, 301, 305, 309, 313, 317,
321, 325, 329, 333, 337, 341, 345, 349, 353, 357, 361, 365, 369, 373, 377, 381,
385, 389, 393, 397, 401, 405, 409, 413, 417, 421, 425, 429, 433, 437, 441, 445,
449, 453, 457, 461, 465, 469, 473, 477, 481, 485, 489, 493, 497, 501, 505, 509,
513, 517, 521, 525, 529, 533, 537, 541, 545, 549, 553, 557, 561, 565, 569, 573,
577, 581, 585, 589, 593, 597, 601, 605, 609, 613, 617, 621, 625, 629, 633, 637,
641, 645, 649, 653, 657, 661, 665, 669, 673, 677, 681, 685, 689, 693, 697, 701,
705, 709, 713, 717, 721, 725, 729, 733, 737, 741, 745, 749, 753, 757, 761, 765,
769, 773, 777, 781, 785, 789, 793, 797, 801, 805, 809, 813, 817, 821, 825, 829,
833, 837, 841, 845, 849, 853, 857, 861, 865, 869, 873, 877, 881, 885, 889, 893,
897, 901, 905, 909, 913, 917, 921, 925, 929, 933, 937, 941, 945, 949, 953, 957,
961, 965, 969, 973, 977, 981, 985, 989, 993, 997, 1001, 1005, 1009, 1013, 1017, 1023
   );

pragma End if;

   COMPRESSED_EVENT_X_TABLE : T_COMPRESSED_EVENT_POSITION_TABLE;
   COMPRESSED_EVENT_Y_TABLE : T_COMPRESSED_EVENT_POSITION_TABLE;
   EVENT_X_TABLE : T_EVENT_POSITION_TABLE;
   EVENT_Y_TABLE : T_EVENT_POSITION_TABLE;


   -- Data types used to store energy conversion tables.

   subtype T_ENERGY_TABLE is IASW_COMMON.T_ARRAY_OF_WORDS(0..4095);

   -- Default energy conversion tables in compressed form.
   DEFAULT_COMPRESSED_ENERGY_TABLE :
      constant IASW_COMMON.T_COMPRESSED_ENERGY_TABLE := (
      -- E(i=1..256) = 4095*sqr(i/256);
      -- new since ver 1.58
   0,    3,    6,    9,   12,   15,   18,   21,   24,   27,   30,   33,   36,   39,   42,   45,
  48,   51,   54,   57,   60,   63,   66,   69,   72,   75,   78,   81,   84,   87,   90,   93,
  96,  100,  104,  108,  112,  116,  120,  124,  128,  132,  136,  140,  144,  148,  152,  156,
 160,  164,  168,  172,  176,  180,  184,  188,  192,  196,  200,  204,  208,  212,  216,  220,
 224,  228,  232,  236,  240,  244,  248,  252,  256,  260,  264,  268,  272,  276,  280,  284,
 288,  292,  296,  300,  304,  308,  312,  316,  320,  324,  328,  332,  336,  340,  344,  348,
 352,  356,  360,  364,  368,  372,  376,  380,  384,  388,  392,  396,  400,  404,  408,  412,
 416,  421,  426,  431,  436,  441,  446,  451,  456,  461,  466,  471,  476,  481,  486,  491,
 496,  504,  512,  520,  528,  536,  544,  552,  560,  568,  576,  584,  592,  600,  608,  616,
 624,  632,  640,  648,  656,  664,  672,  680,  688,  696,  704,  712,  720,  728,  736,  744,
 752,  765,  778,  791,  804,  817,  830,  843,  856,  869,  882,  895,  908,  921,  934,  947,
 960,  973,  986,  999, 1012, 1025, 1038, 1051, 1064, 1077, 1090, 1103, 1116, 1129, 1142, 1155,
1168, 1186, 1204, 1222, 1240, 1258, 1276, 1294, 1312, 1330, 1348, 1366, 1384, 1402, 1420, 1438,
1456, 1474, 1492, 1510, 1528, 1546, 1564, 1582, 1600, 1618, 1636, 1654, 1672, 1690, 1708, 1726,
1744, 1808, 1872, 1936, 2000, 2064, 2128, 2192, 2256, 2320, 2384, 2448, 2512, 2576, 2640, 2704,
2768, 2832, 2896, 2960, 3024, 3088, 3152, 3216, 3280, 3344, 3408, 3472, 3536, 3600, 3664, 4095);

   -- Current energy conversion tables.
   COMPRESSED_ENERGY_TABLE : IASW_COMMON.T_COMPRESSED_ENERGY_TABLE;
   ENERGY_TABLE            : T_ENERGY_TABLE;

   ----------------------------------------------------------------
   -- Local procedures, made global because of possible patching --
   ----------------------------------------------------------------

   procedure ACKNOWLEDGE_OF_GREY_FILTER_CHANGE(GF : in CCT.T_UNSIGNED_16);
   function NO_DATA_IN_THE_BUFFER return BOOLEAN;
   procedure SET_SC_TM_DATA_HEADER(
      TM_BUFF  : out CIT.T_SCIENCE_DATA_FIELD;
      TYP      : in IASW_COMMON.T_PACKET_TYPE;
      STYP     : in IASW_COMMON.T_PACKET_SUBTYPE;
      OBT      : in CIT.T_OBT );
   procedure SC_TM_OUT (SC_BUF : in out T_OUT_BUF);
   procedure PUT_BYTE_TO_TM_DATA(
      BYTE_DATA  : in CCT.T_UNSIGNED_8;
      DATA_ADDR  : in SYSTEM.ADDRESS;
      DATA_INDEX : in out INTEGER );
   function GET_DELTA_TIME(
      CC2, TC2, CC1, TC1 : in CCT.T_UNSIGNED_16
   ) return CIT.T_U32_BITS;
   function TRUNC16(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16;
   function MUL8(X : in CIT.T_U32_BITS) return CIT.T_U32_BITS;
   function DIV4(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16;
   function DIV1024(X : in CIT.T_U32_BITS) return CCT.T_UNSIGNED_16;
   function CLEAR10BITS(X : in CCT.T_UNSIGNED_16) return CCT.T_UNSIGNED_16;
   function COORD_8BIT(
      X : in CCT.T_UNSIGNED_16;
      EVENT_POSITION_TABLE : in T_EVENT_POSITION_TABLE
   ) return CCT.T_UNSIGNED_8;
   procedure FIXBUGG1_INIT;
   procedure FIXBUGG1_DO(EVENT : in out IASW_COMMON.T_EVENT);
   procedure FIXBUGG1_BACK;
   procedure CR_INIT;
   procedure CR_OUT;
   procedure CR_EVENT(
      CC, TC : in CCT.T_UNSIGNED_16;
      DET    : in IASW_COMMON.T_DUMMY_EVENT;
      GFI    : in IASW_GF.T_GREY_FILTER_INDICATOR );
   procedure REQUEST_FOR_SCF_SWITCHING( RQ : in CCT.T_UNSIGNED_16 );
   procedure SCF_UNDEFINED;
   procedure SCF_TEST;
   procedure SCF_FULL;
   procedure SCF_RESTRICTED;
   procedure SCF_SPECTRAL_TIMING;
   procedure PUT_W6_TO_TM_DATA(
      D : in CCT.T_UNSIGNED_6;
      A : in SYSTEM.ADDRESS;
      I : in INTEGER;
      N : out INTEGER );
   procedure SCF_TIMING;
   procedure CLEAR_SPECTRUM;
   procedure PUT4(
      A : in SYSTEM.ADDRESS;
      X : in CCT.T_UNSIGNED_4;
      I : INTEGER );
   procedure PUT8(
      A : in SYSTEM.ADDRESS;
      X : in CCT.T_UNSIGNED_8;
      I : INTEGER );
   procedure PACK_SPECTRUM(TM_DATA_ADDR : in SYSTEM.ADDRESS);
   procedure SCF_SPECTRUM;
   procedure SCF_CALIBRATION;
   procedure SCF_DIAGNOSTIC;
   procedure GET_AND_COMPRESS_EVENTS;
   procedure SET_APID_FOR_THE_FORMAT(SCF : in IASW_MODES.T_SCIENCE_FORMAT);
   procedure SWITCH_TO_FORMAT(
      SCF : IASW_MODES.T_SCIENCE_FORMAT;
      EVS : IASW_COMMON.T_EVENT_SIZE );

   function IS_ENERGY_TABLE_VALID(
      DATA      : in IASW_COMMON.T_COMPRESSED_ENERGY_TABLE
   ) return BOOLEAN;
   procedure DECOMPRESS_ENERGY_TABLE(
      DATA      : in  IASW_COMMON.T_COMPRESSED_ENERGY_TABLE;
      TABLE     : out T_ENERGY_TABLE );
   procedure SET_ENERGY_TABLE (
      PARAM_VAL : in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN : in IASW_COMMON.T_SW_PARAM_LEN );
   procedure EXEC_ENERGY_TABLE;

   function IS_EVENT_POSITION_TABLE_VALID(
      DATA      : in T_COMPRESSED_EVENT_POSITION_TABLE
   ) return BOOLEAN;
   procedure DECOMPRESS_EVENT_POSITION_TABLE(
      DATA      : in  T_COMPRESSED_EVENT_POSITION_TABLE;
      TABLE     : out T_EVENT_POSITION_TABLE );
   procedure SET_EVENT_POSITION_TABLE (
      PARAM_VAL : in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN : in IASW_COMMON.T_SW_PARAM_LEN;
      COMPRESSED_EVENT_POSITION_TABLE : out T_COMPRESSED_EVENT_POSITION_TABLE);
   procedure EXEC_EVENT_POSITION_TABLE(
      COMPRESSED_EVENT_POSITION_TABLE : in out T_COMPRESSED_EVENT_POSITION_TABLE;
      EVENT_POSITION_TABLE : out T_EVENT_POSITION_TABLE;
      DEFAULT_COMPRESSED_EVENT_POSITION_TABLE : in T_COMPRESSED_EVENT_POSITION_TABLE);

   -----------------------
   -- Global procedures --
   -----------------------

   procedure TASK_SC_OUT;
   procedure START_DATA_TAKING;
   procedure SCTM_STOP(STOP_TYPE : in T_SCTM_STOP_TYPE);
   function STOPPED return BOOLEAN;
   procedure SET_SW_PARAM(
      PARAM_ID : in IASW_COMMON.T_FID;
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN);
   procedure GET_SW_PARAM(
      PARAM_ID : in  IASW_COMMON.T_FID;
      PARAM_VAL: out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN: out IASW_COMMON.T_SW_PARAM_LEN);
   procedure BCP1_TICK;
   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE);
   procedure CONSTRUCTOR;

end IASW_SC_OUT;
