
with CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;

package IASW_TIME is

   package CIT renames CSSW_IF_TYPES;
   package CCT renames CSSW_COMMON_TYPES;

   --------------------------------------------------------------------
   -- Local types and data, made global because of possible patching --
   --------------------------------------------------------------------

   BCP1_OBT          : CIT.T_OBT;    -- OBT in the moment of last BCP1.
   DFEE_HK_CC_STORED : CCT.T_UNSIGNED_16; -- DFEE HK Cycle Counter.
   DFEE_OBT_STORED   : CIT.T_OBT;    -- OBT correlated with DFEE_HK_CC_STORED.
   TIME_SEMAPHORE    : CCT.T_UNSIGNED_16 := 0;

   -----------------------
   -- Global procedures --
   -----------------------

   procedure BCP1_TICK;
   -- This should be called just at the reception of BCP1.  Its task is
   -- to store OBT corresponding to BCP1.

   procedure SYNCHRONIZE_DFEE_WITH_OBT(NEW_DFEE_HK_CC : in CCT.T_UNSIGNED_16);
   -- This should be called just after receiving HK Block from DFEE.

   function GET_OBT_OF_EVENT(
      DFEE_HK_CC        : in CCT.T_UNSIGNED_16;
      DFEE_TIME_COUNTER : in CCT.T_UNSIGNED_16
   ) return CIT.T_OBT;
   -- This function returns time of a given event in OBT scale.

   function SEC_TO_OBT(
      SEC  : in LONG_INTEGER;
      FSEC : in CCT.T_UNSIGNED_16 := 0
   ) return CIT.T_OBT;
   -- This function returns number of seconds in the format of OBT.
   -- Argument SEC can be of any value - positive, zero and negative.
   -- FSEC (fraction of sec) is 16 bits following the full sec position.
   -- it is always interpreted as positive.

   function "+"( A,B : in CIT.T_OBT ) return CIT.T_OBT;
   -- Intended use:  A=OBT, B=offset, e.g SEC_TO_OBT(10)

   function "<"( A,B : in CIT.T_OBT ) return BOOLEAN;
   -- Comparing of two OBTs.
   -- Example of the usage:
   -- if (ECLIPSE_EXIT_OBT + SEC_TO_OBT(10)) < DPE_CSSW.GET_CURRENT_OBT then

   procedure CONSTRUCTOR;

end IASW_TIME;
